# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A library of classes representing computations in a deserialized form."""

import abc
from collections.abc import Iterable, Iterator
import enum
import typing
from typing import Optional, Union
import zlib

from federated_language.common_libs import py_typecheck
from federated_language.common_libs import structure
from federated_language.compiler import array
from federated_language.compiler import intrinsic_defs
from federated_language.proto import computation_pb2
from federated_language.types import computation_types
from federated_language.types import placements
from federated_language.types import type_analysis
from federated_language.types import typed_object
import numpy as np

from google.protobuf import any_pb2


def _check_computation_has_field(
    computation_pb: computation_pb2.Computation, field: str
):
  if not computation_pb.HasField(field):
    raise ValueError(
        f'Expected `computation_pb` to have the field "{field}", found'
        f' {computation_pb}.'
    )


class UnexpectedBlockError(TypeError):

  def __init__(
      self,
      expected: type['ComputationBuildingBlock'],
      actual: 'ComputationBuildingBlock',
  ):
    message = f'Expected block of kind {expected}, found block {actual}'
    super().__init__(message)
    self.actual = actual
    self.expected = expected


class ComputationBuildingBlock(typed_object.TypedObject, abc.ABC):
  """The abstract base class for abstractions in the internal language.

  Instances of this class correspond roughly one-to-one to the abstractions
  defined in the `Computation` message in the `computation.proto`, and are
  intended primarily for the ease of manipulating the abstract syntax trees
  (AST) of federated computations as they are transformed by the compiler
  pipeline to mold into the needs of a particular execution backend. The only
  abstraction that does not have a dedicated Python equivalent is a section
  of backend-specific code (it's represented by
  `federated_language.framework.CompiledComputation`).
  """

  def __init__(self, type_signature):
    """Constructs a computation building block with the given type.

    Args:
      type_signature: An instance of types.Type, or something convertible to it
        via types.to_type().
    """
    self._type_signature = computation_types.to_type(type_signature)
    self._hash = None
    self._proto = None

  @classmethod
  def from_proto(
      cls, computation_pb: computation_pb2.Computation
  ) -> 'ComputationBuildingBlock':
    """Returns a `ComputationBuildingBlock` for the `computation_pb`."""
    computation_oneof = computation_pb.WhichOneof('computation')
    if computation_oneof == 'block':
      return Block.from_proto(computation_pb)
    elif computation_oneof == 'call':
      return Call.from_proto(computation_pb)
    elif computation_oneof == 'data':
      return Data.from_proto(computation_pb)
    elif computation_oneof == 'intrinsic':
      return Intrinsic.from_proto(computation_pb)
    elif computation_oneof == 'lambda':
      return Lambda.from_proto(computation_pb)
    elif computation_oneof == 'literal':
      return Literal.from_proto(computation_pb)
    elif computation_oneof == 'placement':
      return Placement.from_proto(computation_pb)
    elif computation_oneof == 'reference':
      return Reference.from_proto(computation_pb)
    elif computation_oneof == 'selection':
      return Selection.from_proto(computation_pb)
    elif computation_oneof == 'struct':
      return Struct.from_proto(computation_pb)
    elif computation_oneof == 'tensorflow':
      return CompiledComputation.from_proto(computation_pb)
    elif computation_oneof == 'xla':
      return CompiledComputation.from_proto(computation_pb)
    else:
      raise NotImplementedError(
          f'Unexpected computation found: {computation_oneof}.'
      )

  @abc.abstractmethod
  def to_proto(self):
    """Returns a `computation_pb2.Computation` for this building block."""
    raise NotImplementedError

  @property
  def type_signature(self) -> computation_types.Type:
    return self._type_signature

  @abc.abstractmethod
  def children(self) -> Iterator['ComputationBuildingBlock']:
    """Returns an iterator yielding immediate child building blocks."""
    raise NotImplementedError

  def compact_representation(self):
    """Returns the compact string representation of this building block."""
    return _string_representation(self, formatted=False)

  def formatted_representation(self):
    """Returns the formatted string representation of this building block."""
    return _string_representation(self, formatted=True)

  def structural_representation(self):
    """Returns the structural string representation of this building block."""
    return _structural_representation(self)

  def __str__(self):
    return self.compact_representation()

  @abc.abstractmethod
  def __repr__(self):
    raise NotImplementedError


class Reference(ComputationBuildingBlock):
  """A reference to a name defined earlier in the internal language.

  Names are defined by lambda expressions (which have formal named parameters),
  and block structures (which can have one or more locals). The reference
  construct is used to refer to those parameters or locals by a string name.
  The usual hiding rules apply. A reference binds to the closest definition of
  the given name in the most deeply nested surrounding lambda or block.

  A concise notation for a reference to name `foo` is `foo`. For example, in
  a lambda expression `(x -> f(x))` there are two references, one to `x` that
  is defined as the formal parameter of the lambda epxression, and one to `f`
  that must have been defined somewhere in the surrounding context.
  """

  def __init__(self, name: str, type_signature: object, context=None):
    """Creates a reference to 'name' of type 'type_signature' in context 'context'.

    Args:
      name: The name of the referenced entity.
      type_signature: The type spec of the referenced entity.
      context: The optional context in which the referenced entity is defined.
        This class does not prescribe what Python type the 'context' needs to be
        and merely exposes it as a property (see below). The only requirement is
        that the context implements str() and repr().

    Raises:
      TypeError: if the arguments are of the wrong types.
    """
    super().__init__(type_signature)
    self._name = name
    self._context = context

  @classmethod
  def from_proto(
      cls, computation_pb: computation_pb2.Computation
  ) -> 'Reference':
    """Returns a `Reference` for the `computation_pb`."""
    _check_computation_has_field(computation_pb, 'reference')

    type_signature = computation_types.Type.from_proto(computation_pb.type)
    return Reference(computation_pb.reference.name, type_signature)

  def to_proto(self) -> computation_pb2.Computation:
    """Returns a `computation_pb2.Computation` for this building block."""
    if self._proto is None:
      type_pb = self._type_signature.to_proto()
      reference_pb = computation_pb2.Reference(name=self._name)
      self._proto = computation_pb2.Computation(
          type=type_pb,
          reference=reference_pb,
      )
    return self._proto

  def children(self) -> Iterator[ComputationBuildingBlock]:
    del self
    return iter(())

  @property
  def name(self) -> str:
    return self._name

  @property
  def context(self):
    return self._context

  def __eq__(self, other: object) -> bool:
    if self is other:
      return True
    elif not isinstance(other, Reference):
      return NotImplemented
    # IMPORTANT: References are only equal to each other if they are the same
    # object because two references with the same `name` are different if they
    # are in different locations within the same scope, in different scopes, or
    # in different contexts.
    return False

  def __hash__(self):
    if self._hash is None:
      self._hash = hash((self._name, self._type_signature))
    return self._hash

  def __repr__(self):
    if self._context is not None:
      return (
          f'Reference({self._name!r}, {self._type_signature!r},'
          f' {self._context!r})'
      )
    else:
      return f'Reference({self._name!r}, {self._type_signature!r})'


class Selection(ComputationBuildingBlock):
  """A selection by name or index from a struct-typed value in the language.

  The concise syntax for selections is `foo.bar` (selecting a named `bar` from
  the value of expression `foo`), and `foo[n]` (selecting element at index `n`
  from the value of `foo`).
  """

  def __init__(
      self,
      source: ComputationBuildingBlock,
      name: Optional[str] = None,
      index: Optional[int] = None,
  ):
    """A selection from 'source' by a string or numeric 'name_or_index'.

    Exactly one of 'name' or 'index' must be specified (not None).

    Args:
      source: The source value to select from (an instance of
        ComputationBuildingBlock).
      name: A string name of the element to be selected.
      index: A numeric index of the element to be selected.

    Raises:
      TypeError: if arguments are of the wrong types.
      ValueError: if the name is empty or index is negative, or the name/index
        is not compatible with the type signature of the source, or neither or
        both are defined (not None).
    """
    source_type = source.type_signature
    # TODO: b/224484886 - Downcasting to all handled types.
    source_type = typing.cast(Union[computation_types.StructType], source_type)
    if not isinstance(source_type, computation_types.StructType):
      raise TypeError(
          'Expected the source of selection to be a `StructType`, '
          'instead found it to be of type {}.'.format(source_type)
      )
    if name is not None and index is not None:
      raise ValueError(
          'Cannot simultaneously specify a name and an index, choose one.'
      )
    if name is not None:
      if not name:
        raise ValueError('The name of the selected element cannot be empty.')
      # Normalize, in case we are dealing with a Unicode type or some such.
      name = str(name)
      if name not in source_type.fields():
        raise ValueError(
            f'Error selecting named field `{name}` from type `{source_type}`, '
            f'whose only named fields are {source_type.fields()}.'
        )
      type_signature = source_type[name]
    elif index is not None:
      length = len(source_type)
      if index < 0 or index >= length:
        raise ValueError(
            f'The index `{index}` does not fit into the valid range in the '
            f'struct type: 0..{length}'
        )
      type_signature = source_type[index]
    else:
      raise ValueError(
          'Must define either a name or index, and neither was specified.'
      )
    super().__init__(type_signature)
    self._source = source
    self._name = name
    self._index = index

  @classmethod
  def from_proto(
      cls, computation_pb: computation_pb2.Computation
  ) -> 'Selection':
    """Returns a `Selection` for the `computation_pb`."""
    _check_computation_has_field(computation_pb, 'selection')

    source = ComputationBuildingBlock.from_proto(
        computation_pb.selection.source
    )
    return Selection(source, index=computation_pb.selection.index)

  def to_proto(self) -> computation_pb2.Computation:
    """Returns a `computation_pb2.Computation` for this building block."""
    if self._proto is None:
      type_pb = self._type_signature.to_proto()
      source_pb = self._source.to_proto()
      selection_pb = computation_pb2.Selection(
          source=source_pb,
          index=self.as_index(),
      )
      return computation_pb2.Computation(
          type=type_pb,
          selection=selection_pb,
      )
    return self._proto

  def children(self) -> Iterator[ComputationBuildingBlock]:
    yield self._source

  @property
  def source(self) -> ComputationBuildingBlock:
    return self._source

  @property
  def name(self) -> Optional[str]:
    return self._name

  @property
  def index(self) -> Optional[int]:
    return self._index

  def as_index(self) -> int:
    if self._index is not None:
      return self._index
    else:
      names = [n for n, _ in self._source.type_signature.items()]  # pytype: disable=attribute-error
      return names.index(self._name)

  def __eq__(self, other: object) -> bool:
    if self is other:
      return True
    elif not isinstance(other, Selection):
      return NotImplemented
    return (
        self._source,
        self._name,
        self._index,
    ) == (
        other._source,
        other._name,
        other._index,
    )

  def __hash__(self):
    if self._hash is None:
      self._hash = hash((self._source, self._name, self._index))
    return self._hash

  def __repr__(self):
    if self._name is not None:
      return f'Selection({self._source!r}, name={self._name!r})'
    else:
      return f'Selection({self._source!r}, index={self._index!r})'


class Struct(ComputationBuildingBlock, structure.Struct):
  """A struct with named or unnamed elements in the internal language.

  The concise notation for structs is `<name_1=value_1, ...., name_n=value_n>`
  for structs with named elements, `<value_1, ..., value_n>` for structs with
  unnamed elements, or a mixture of these for structs with some named and some
  unnamed elements, where `name_k` are the names, and `value_k` are the value
  expressions.

  For example, a lambda expression that applies `fn` to elements of 2-structs
  pointwise could be represented as `(arg -> <fn(arg[0]),fn(arg[1])>)`.
  """

  def __init__(self, elements, container_type=None):
    """Constructs a struct from the given list of elements.

    Args:
      elements: The elements of the struct, supplied as a list of (name, value)
        pairs, where 'name' can be None in case the corresponding element is not
        named and only accessible via an index (see also `structure.Struct`).
      container_type: An optional Python container type to associate with the
        struct.

    Raises:
      TypeError: if arguments are of the wrong types.
    """

    # Not using super() here and below, as the two base classes have different
    # signatures of their constructors, and the struct implementation
    # of selection interfaces should override that in the generic class 'Value'
    # to favor simplified expressions where simplification is possible.
    def _map_element(e):
      """Returns a named or unnamed element."""
      if isinstance(e, ComputationBuildingBlock):
        return (None, e)
      elif py_typecheck.is_name_value_pair(
          e, value_type=ComputationBuildingBlock
      ):
        if e[0] is not None and not e[0]:
          raise ValueError('Unexpected struct element with empty string name.')
        return (e[0], e[1])
      else:
        raise TypeError('Unexpected struct element: {}.'.format(e))

    elements = [_map_element(e) for e in elements]
    element_pairs = [
        ((e[0], e[1].type_signature) if e[0] else e[1].type_signature)
        for e in elements
    ]

    if container_type is None:
      type_signature = computation_types.StructType(element_pairs)
    else:
      type_signature = computation_types.StructWithPythonType(
          element_pairs, container_type
      )
    ComputationBuildingBlock.__init__(self, type_signature)
    structure.Struct.__init__(self, elements)
    self._type_signature = type_signature

  @classmethod
  def from_proto(cls, computation_pb: computation_pb2.Computation) -> 'Struct':
    """Returns a `Struct` for the `computation_pb`."""
    _check_computation_has_field(computation_pb, 'struct')
    elements = []
    for element_pb in computation_pb.struct.element:
      if element_pb.name:
        name = element_pb.name
      else:
        name = None
      element = ComputationBuildingBlock.from_proto(element_pb.value)
      elements.append((name, element))
    return Struct(elements)

  def to_proto(self) -> computation_pb2.Computation:
    """Returns a `computation_pb2.Computation` for this building block."""
    if self._proto is None:
      type_pb = self._type_signature.to_proto()
      element_pbs = []
      for name, element in self.items():
        value_pb = element.to_proto()
        element_pb = computation_pb2.Struct.Element(
            name=name,
            value=value_pb,
        )
        element_pbs.append(element_pb)
      struct_pb = computation_pb2.Struct(element=element_pbs)
      self._proto = computation_pb2.Computation(
          type=type_pb,
          struct=struct_pb,
      )
    return self._proto

  @property
  def type_signature(self) -> computation_types.StructType:
    return self._type_signature

  def children(self) -> Iterator[ComputationBuildingBlock]:
    return (element for _, element in structure.iter_elements(self))

  def items(self) -> Iterator[tuple[Optional[str], ComputationBuildingBlock]]:
    return structure.iter_elements(self)

  def __eq__(self, other: object) -> bool:
    if self is other:
      return True
    elif not isinstance(other, Struct):
      return NotImplemented
    if self._type_signature != other._type_signature:
      return False
    return structure.Struct.__eq__(self, other)

  def __hash__(self):
    if self._hash is None:
      self._hash = hash((
          structure.Struct.__hash__(self),
          self._type_signature,
      ))
    return self._hash

  def __repr__(self):
    elements = list(self.items())
    return f'Struct({elements!r})'


class Call(ComputationBuildingBlock):
  """A representation of a function invocation in the internal language.

  The call construct takes an argument struct with two elements, the first being
  the function to invoke (represented as a computation with a functional result
  type), and the second being the argument to feed to that function. Typically,
  the function is either an intrinsic, or a lambda expression.

  The concise notation for calls is `foo(bar)`, where `foo` is the function,
  and `bar` is the argument.
  """

  def __init__(
      self,
      fn: ComputationBuildingBlock,
      arg: Optional[ComputationBuildingBlock] = None,
  ):
    """Creates a call to 'fn' with argument 'arg'.

    Args:
      fn: A value of a functional type that represents the function to invoke.
      arg: The optional argument, present iff 'fn' expects one, of a type that
        matches the type of 'fn'.

    Raises:
      TypeError: if the arguments are of the wrong types.
    """
    function_type = fn.type_signature
    # TODO: b/224484886 - Downcasting to all handled types.
    function_type = typing.cast(
        Union[computation_types.FunctionType], function_type
    )
    if not isinstance(function_type, computation_types.FunctionType):
      raise TypeError(
          'Expected `fn` to have a `federated_language.FunctionType`, found'
          f' {function_type}.'
      )
    parameter_type = function_type.parameter
    if parameter_type is not None:
      if arg is None:
        raise TypeError(
            f'Expected `arg` to be of type {parameter_type}, found None.'
        )
      elif not parameter_type.is_assignable_from(arg.type_signature):
        raise TypeError(
            f'Expected `arg` to be of type {parameter_type}, found an'
            f' incompatible type {arg.type_signature}.'
        )
    else:
      if arg is not None:
        raise TypeError(f'Expected `arg` to be None, found {arg}.')
    super().__init__(function_type.result)
    self._function = fn
    self._argument = arg

  @classmethod
  def from_proto(cls, computation_pb: computation_pb2.Computation) -> 'Call':
    """Returns a `Call` for the `computation_pb`."""
    _check_computation_has_field(computation_pb, 'call')

    fn = ComputationBuildingBlock.from_proto(computation_pb.call.function)
    arg_proto = computation_pb.call.argument
    if arg_proto.WhichOneof('computation') is not None:
      arg = ComputationBuildingBlock.from_proto(arg_proto)
    else:
      arg = None
    return Call(fn, arg)

  def to_proto(self) -> computation_pb2.Computation:
    """Returns a `computation_pb2.Computation` for this building block."""
    if self._proto is None:
      type_pb = self._type_signature.to_proto()
      function_pb = self._function.to_proto()
      if self._argument is not None:
        argument_pb = self._argument.to_proto()
      else:
        argument_pb = None
      call_pb = computation_pb2.Call(
          function=function_pb,
          argument=argument_pb,
      )
      self._proto = computation_pb2.Computation(
          type=type_pb,
          call=call_pb,
      )
    return self._proto

  def children(self) -> Iterator[ComputationBuildingBlock]:
    yield self._function
    if self._argument is not None:
      yield self._argument

  @property
  def function(self) -> ComputationBuildingBlock:
    return self._function

  @property
  def argument(self) -> Optional[ComputationBuildingBlock]:
    return self._argument

  def __eq__(self, other: object) -> bool:
    if self is other:
      return True
    elif not isinstance(other, Call):
      return NotImplemented
    return (
        self._function,
        self._argument,
    ) == (
        other._function,
        other._argument,
    )

  def __hash__(self):
    if self._hash is None:
      self._hash = hash((self._function, self._argument))
    return self._hash

  def __repr__(self):
    if self._argument is not None:
      return f'Call({self._function!r}, {self._argument!r})'
    else:
      return f'Call({self._function!r})'


class Lambda(ComputationBuildingBlock):
  """A representation of a lambda expression in the internal language.

  A lambda expression consists of a string formal parameter name, and a result
  expression that can contain references by name to that formal parameter. A
  concise notation for lambdas is `(foo -> bar)`, where `foo` is the name of
  the formal parameter, and `bar` is the result expression.
  """

  def __init__(
      self,
      parameter_name: Optional[str],
      parameter_type: Optional[object],
      result: ComputationBuildingBlock,
  ):
    """Creates a lambda expression.

    Args:
      parameter_name: The (string) name of the parameter accepted by the lambda.
        This name can be used by Reference() instances in the body of the lambda
        to refer to the parameter. Note that an empty parameter name shall be
        treated as equivalent to no parameter.
      parameter_type: The type of the parameter, an instance of types.Type or
        something convertible to it by types.to_type().
      result: The resulting value produced by the expression that forms the body
        of the lambda. Must be an instance of ComputationBuildingBlock.

    Raises:
      TypeError: if the arguments are of the wrong types.
    """
    if not parameter_name:
      parameter_name = None
    if (parameter_name is None) != (parameter_type is None):
      raise TypeError(
          'A lambda expression must have either a valid parameter name and '
          'type or both parameter name and type must be `None`. '
          '`parameter_name` was {} but `parameter_type` was {}.'.format(
              parameter_name, parameter_type
          )
      )
    if parameter_name is not None:
      parameter_type = computation_types.to_type(parameter_type)
    type_signature = computation_types.FunctionType(
        parameter_type, result.type_signature
    )
    super().__init__(type_signature)
    self._parameter_name = parameter_name
    self._parameter_type = parameter_type
    self._result = result
    self._type_signature = type_signature

  @classmethod
  def from_proto(cls, computation_pb: computation_pb2.Computation) -> 'Lambda':
    """Returns a `Lambda` for the `computation_pb`."""
    _check_computation_has_field(computation_pb, 'lambda')

    fn_pb: computation_pb2.Lambda = getattr(computation_pb, 'lambda')
    if computation_pb.type.function.HasField('parameter'):
      parameter_type = computation_types.Type.from_proto(
          computation_pb.type.function.parameter
      )
    else:
      parameter_type = None
    result = ComputationBuildingBlock.from_proto(fn_pb.result)
    return Lambda(fn_pb.parameter_name, parameter_type, result)

  def to_proto(self) -> computation_pb2.Computation:
    """Returns a `computation_pb2.Computation` for this building block."""
    if self._proto is None:
      type_pb = self._type_signature.to_proto()
      result_pb = self._result.to_proto()
      fn_pb = computation_pb2.Lambda(
          parameter_name=self._parameter_name,
          result=result_pb,
      )
      # We are unpacking the lambda argument here because `lambda` is a reserved
      # keyword in Python, but it is also the name of the parameter for a
      # `computation_pb2.Computation`.
      self._proto = computation_pb2.Computation(
          type=type_pb, **{'lambda': fn_pb}
      )
    return self._proto

  @property
  def type_signature(self) -> computation_types.FunctionType:
    return self._type_signature

  def children(self) -> Iterator[ComputationBuildingBlock]:
    yield self._result

  @property
  def parameter_name(self) -> Optional[str]:
    return self._parameter_name

  @property
  def parameter_type(self) -> Optional[computation_types.Type]:
    return self._parameter_type

  @property
  def result(self) -> ComputationBuildingBlock:
    return self._result

  def __eq__(self, other: object) -> bool:
    if self is other:
      return True
    elif not isinstance(other, Lambda):
      return NotImplemented
    return (
        self._parameter_name,
        self._parameter_type,
        self._result,
    ) == (
        other._parameter_name,
        other._parameter_type,
        other._result,
    )

  def __hash__(self):
    if self._hash is None:
      self._hash = hash((
          self._parameter_name,
          self._parameter_type,
          self._result,
      ))
    return self._hash

  def __repr__(self) -> str:
    return (
        f'Lambda({self._parameter_name!r}, {self._parameter_type!r},'
        f' {self._result!r})'
    )


class Block(ComputationBuildingBlock):
  """A representation of a block of code in the internal language.

  A block is a syntactic structure that consists of a sequence of local name
  bindings followed by a result. The bindings are interpreted sequentially,
  with bindings later in the sequence in the scope of those listed earlier,
  and the result in the scope of the entire sequence. The usual hiding rules
  apply.

  An informal concise notation for blocks is the following, with `name_k`
  representing the names defined locally for the block, `value_k` the values
  associated with them, and `result` being the expression that reprsents the
  value of the block construct.

  ```
  let name_1=value_1, name_2=value_2, ..., name_n=value_n in result
  ```

  Blocks are technically a redundant abstraction, as they can be equally well
  represented by lambda expressions. A block of the form `let x=y in z` is
  roughly equivalent to `(x -> z)(y)`. Although redundant, blocks have a use
  as a way to reduce computation ASTs to a simpler, less nested and more
  readable form, and are helpful in AST transformations as a mechanism that
  prevents possible naming conflicts.

  An example use of a block expression to flatten a nested structure below:

  ```
  z = federated_sum(federated_map(x, federated_broadcast(y)))
  ```

  An equivalent form in a more sequential notation using a block expression:
  ```
  let
    v1 = federated_broadcast(y),
    v2 = federated_map(x, v1)
  in
    federated_sum(v2)
  ```
  """

  def __init__(
      self,
      local_symbols: Iterable[tuple[str, ComputationBuildingBlock]],
      result: ComputationBuildingBlock,
  ):
    """Creates a block of federated_language code.

    Args:
      local_symbols: The list of one or more local declarations, each of which
        is a 2-tuple (name, value), with 'name' being the string name of a local
        symbol being defined, and 'value' being the instance of
        ComputationBuildingBlock, the output of which will be locally bound to
        that name.
      result: An instance of ComputationBuildingBlock that computes the result.

    Raises:
      TypeError: if the arguments are of the wrong types.
    """
    updated_locals = []
    for index, element in enumerate(local_symbols):
      if (
          not isinstance(element, tuple)
          or (len(element) != 2)
          or not isinstance(element[0], str)
      ):
        raise TypeError(
            'Expected the locals to be a list of 2-element structs with string '
            'name as their first element, but this is not the case for the '
            'local at position {} in the sequence: {}.'.format(index, element)
        )
      name = element[0]
      value = element[1]
      updated_locals.append((name, value))
    super().__init__(result.type_signature)
    self._local_symbols = updated_locals
    self._result = result

  @classmethod
  def from_proto(cls, computation_pb: computation_pb2.Computation) -> 'Block':
    """Returns a `Block` for the `computation_pb`."""
    _check_computation_has_field(computation_pb, 'block')

    local_symbols = []
    for local_symbol_pb in computation_pb.block.local:
      if local_symbol_pb.name:
        name = local_symbol_pb.name
      else:
        name = None
      symbol = ComputationBuildingBlock.from_proto(local_symbol_pb.value)
      local_symbols.append((name, symbol))
    result = ComputationBuildingBlock.from_proto(computation_pb.block.result)
    return Block(local_symbols, result)

  def to_proto(self) -> computation_pb2.Computation:
    """Returns a `computation_pb2.Computation` for this building block."""
    if self._proto is None:
      type_pb = self._type_signature.to_proto()

      local_symbol_pbs = []
      for name, local_symbol in self._local_symbols:
        value_pb = local_symbol.to_proto()
        local_symbol_pb = computation_pb2.Block.Local(
            name=name,
            value=value_pb,
        )
        local_symbol_pbs.append(local_symbol_pb)
      result_pb = self._result.to_proto()
      block_pb = computation_pb2.Block(
          **{'local': local_symbol_pbs},
          result=result_pb,
      )
      self._proto = computation_pb2.Computation(
          type=type_pb,
          block=block_pb,
      )
    return self._proto

  def children(self) -> Iterator[ComputationBuildingBlock]:
    for _, value in self._local_symbols:
      yield value
    yield self._result

  @property
  def locals(self) -> list[tuple[str, ComputationBuildingBlock]]:
    return list(self._local_symbols)

  @property
  def result(self) -> ComputationBuildingBlock:
    return self._result

  def __eq__(self, other: object) -> bool:
    if self is other:
      return True
    elif not isinstance(other, Block):
      return NotImplemented
    return (
        self._local_symbols,
        self._result,
    ) == (
        other._local_symbols,
        other._result,
    )

  def __hash__(self):
    if self._hash is None:
      self._hash = hash((tuple(self._local_symbols), self._result))
    return self._hash

  def __repr__(self) -> str:
    return f'Block({self._local_symbols!r}, {self._result!r})'


class Intrinsic(ComputationBuildingBlock):
  """A representation of an intrinsic in the internal language.

  An intrinsic is a symbol known to the compiler pipeline, represented
  as a known URI. It generally appears in expressions with a concrete type,
  although all intrinsic are defined with template types. This class does not
  deal with parsing intrinsic URIs and verifying their types, it is only a
  container. Parsing and type analysis are a responsibility of the components
  that manipulate ASTs. See intrinsic_defs.py for the list of known intrinsics.
  """

  def __init__(self, uri: str, type_signature: computation_types.Type):
    """Creates an intrinsic.

    Args:
      uri: The URI of the intrinsic.
      type_signature: A `federated_language.Type`, the type of the intrinsic.

    Raises:
      TypeError: if the arguments are of the wrong types.
    """
    intrinsic_def = intrinsic_defs.uri_to_intrinsic_def(uri)
    if intrinsic_def is not None:
      # NOTE: this is really expensive.
      type_analysis.check_concrete_instance_of(
          type_signature, intrinsic_def.type_signature
      )
    super().__init__(type_signature)
    self._uri = uri

  @classmethod
  def from_proto(
      cls, computation_pb: computation_pb2.Computation
  ) -> 'Intrinsic':
    """Returns a `Intrinsic` for the `computation_pb`."""
    _check_computation_has_field(computation_pb, 'intrinsic')

    type_signature = computation_types.Type.from_proto(computation_pb.type)
    return Intrinsic(computation_pb.intrinsic.uri, type_signature)

  def to_proto(self) -> computation_pb2.Computation:
    """Returns a `computation_pb2.Computation` for this building block."""
    if self._proto is None:
      type_pb = self._type_signature.to_proto()
      intrinsic_pb = computation_pb2.Intrinsic(uri=self._uri)
      self._proto = computation_pb2.Computation(
          type=type_pb,
          intrinsic=intrinsic_pb,
      )
    return self._proto

  def intrinsic_def(self) -> intrinsic_defs.IntrinsicDef:
    intrinsic_def = intrinsic_defs.uri_to_intrinsic_def(self._uri)
    if intrinsic_def is None:
      raise ValueError(
          'Failed to retrieve definition of intrinsic with URI '
          f'`{self._uri}`. Perhaps a definition needs to be added to '
          '`intrinsic_defs.py`?'
      )
    return intrinsic_def

  def children(self) -> Iterator[ComputationBuildingBlock]:
    del self
    return iter(())

  @property
  def uri(self) -> str:
    return self._uri

  def __eq__(self, other: object) -> bool:
    if self is other:
      return True
    elif not isinstance(other, Intrinsic):
      return NotImplemented
    return (
        self._uri,
        self._type_signature,
    ) == (
        other._uri,
        other._type_signature,
    )

  def __hash__(self):
    if self._hash is None:
      self._hash = hash((self._uri, self._type_signature))
    return self._hash

  def __repr__(self) -> str:
    return f'Intrinsic({self._uri!r}, {self._type_signature!r})'


class Data(ComputationBuildingBlock):
  """A representation of data (an input pipeline).

  This class does not deal with parsing data protos and verifying correctness,
  it is only a container. Parsing and type analysis are a responsibility
  or a component external to this module.
  """

  def __init__(self, content: any_pb2.Any, type_signature: object):
    """Creates a representation of data.

    Args:
      content: The proto that characterizes the data.
      type_signature: Either the types.Type that represents the type of this
        data, or something convertible to it by types.to_type().
    """
    super().__init__(type_signature)
    self._content = content

  @classmethod
  def from_proto(cls, computation_pb: computation_pb2.Computation) -> 'Data':
    """Returns a `Data` for the `computation_pb`."""
    _check_computation_has_field(computation_pb, 'data')

    type_signature = computation_types.Type.from_proto(computation_pb.type)
    return Data(computation_pb.data.content, type_signature)

  def to_proto(self) -> computation_pb2.Computation:
    """Returns a `computation_pb2.Computation` for this building block."""
    if self._proto is None:
      type_pb = self._type_signature.to_proto()
      data_pb = computation_pb2.Data(content=self._content)
      self._proto = computation_pb2.Computation(
          type=type_pb,
          data=data_pb,
      )
    return self._proto

  def children(self) -> Iterator[ComputationBuildingBlock]:
    del self
    return iter(())

  @property
  def content(self) -> any_pb2.Any:
    return self._content

  def __eq__(self, other: object) -> bool:
    if self is other:
      return True
    elif not isinstance(other, Data):
      return NotImplemented
    return (
        self._content,
        self._type_signature,
    ) == (
        other._content,
        other._type_signature,
    )

  def __hash__(self):
    if self._hash is None:
      self._hash = hash((str(self._content), self._type_signature))
    return self._hash

  def __repr__(self) -> str:
    return f'Data({self._content!r}, {self._type_signature!r})'


class CompiledComputation(ComputationBuildingBlock):
  """A representation of a fully constructed and serialized computation.

  A compiled computation is one that has not been parsed into constituents, and
  is simply represented as an embedded `Computation` protocol buffer. Whereas
  technically, any computation can be represented and passed around this way,
  this structure is generally only used to represent TensorFlow sections, for
  which otherwise there isn't any dedicated structure.
  """

  def __init__(
      self,
      proto: computation_pb2.Computation,
      name: Optional[str] = None,
      type_signature: Optional[computation_types.Type] = None,
  ):
    """Creates a representation of a fully constructed computation.

    Args:
      proto: An instance of computation_pb2.Computation with the computation
        logic.
      name: An optional string name to associate with this computation, used
        only for debugging purposes. If the name is not specified (None), it is
        autogenerated as a hexadecimal string from the hash of the proto.
      type_signature: An optional type signature to associate with this
        computation rather than the serialized one.
    """
    if type_signature is None:
      type_signature = computation_types.Type.from_proto(proto.type)
    super().__init__(type_signature)
    self._proto = proto
    if name is None:
      name = '{:x}'.format(zlib.adler32(self._proto.SerializeToString()))
    self._name = name

  @classmethod
  def from_proto(
      cls, computation_pb: computation_pb2.Computation
  ) -> 'CompiledComputation':
    """Returns a `CompiledComputation` for the `computation_pb`."""
    return CompiledComputation(computation_pb)

  def to_proto(self) -> computation_pb2.Computation:
    return self._proto

  def children(self) -> Iterator[ComputationBuildingBlock]:
    del self
    return iter(())

  @property
  def name(self) -> str:
    return self._name

  def __eq__(self, other: object) -> bool:
    if self is other:
      return True
    elif not isinstance(other, CompiledComputation):
      return NotImplemented
    return (
        self._proto,
        self._name,
        self._type_signature,
    ) == (
        other._proto,
        other._name,
        other._type_signature,
    )

  def __hash__(self):
    if self._hash is None:
      self._hash = hash((
          self._proto.SerializeToString(),
          self._name,
          self._type_signature,
      ))
    return self._hash

  def __repr__(self) -> str:
    return f'CompiledComputation({self._name!r}, {self._type_signature!r})'


class Placement(ComputationBuildingBlock):
  """A representation of a placement literal in the internal language.

  Currently this can only be `federated_language.SERVER` or
  `federated_language.CLIENTS`.
  """

  def __init__(self, literal: placements.PlacementLiteral):
    """Constructs a new placement instance for the given placement literal.

    Args:
      literal: The placement literal.

    Raises:
      TypeError: if the arguments are of the wrong types.
    """
    super().__init__(computation_types.PlacementType())
    self._literal = literal

  @classmethod
  def from_proto(
      cls, computation_pb: computation_pb2.Computation
  ) -> 'Placement':
    """Returns a `Placement` for the `computation_pb`."""
    _check_computation_has_field(computation_pb, 'placement')

    literal = placements.uri_to_placement_literal(computation_pb.placement.uri)
    return Placement(literal)

  def to_proto(self) -> computation_pb2.Computation:
    """Returns a `computation_pb2.Computation` for this building block."""
    if self._proto is None:
      type_pb = self._type_signature.to_proto()
      placement_pb = computation_pb2.Placement(uri=self._literal.uri)
      self._proto = computation_pb2.Computation(
          type=type_pb,
          placement=placement_pb,
      )
    return self._proto

  def children(self) -> Iterator[ComputationBuildingBlock]:
    del self
    return iter(())

  @property
  def uri(self) -> str:
    return self._literal.uri

  def __eq__(self, other: object) -> bool:
    if self is other:
      return True
    elif not isinstance(other, Placement):
      return NotImplemented
    return self._literal == other._literal

  def __hash__(self):
    if self._hash is None:
      self._hash = hash((self._literal))
    return self._hash

  def __repr__(self) -> str:
    return f'Placement({self._literal.uri!r})'


class Literal(ComputationBuildingBlock):
  """A representation of a literal in the internal language."""

  def __init__(
      self, value: array.Array, type_signature: computation_types.TensorType
  ):
    """Returns an initialized `federated_language.framework.Literal`.

    Args:
      value: The value of the literal.
      type_signature: A `federated_language.TensorType`.

    Raises:
      ValueError: If `value` is not compatible with `type_signature`.
    """
    if (
        isinstance(value, (np.generic, np.ndarray))
        and value.dtype.type is np.str_
    ):
      value = value.astype(np.bytes_)
    elif isinstance(value, str):
      value = value.encode()

    if not array.is_compatible_dtype(value, type_signature.dtype.type):
      raise ValueError(
          f"Expected '{value}' to be compatible with"
          f" '{type_signature.dtype.type}'."
      )

    if not array.is_compatible_shape(value, type_signature.shape):
      raise ValueError(
          f"Expected '{value}' to be compatible with '{type_signature.shape}'."
      )

    super().__init__(type_signature)
    self._value = value
    self._type_signature = type_signature
    self._hash = None

  @classmethod
  def from_proto(cls, computation_pb: computation_pb2.Computation) -> 'Literal':
    """Returns a `Literal` for the `computation_pb`."""
    _check_computation_has_field(computation_pb, 'literal')

    value = array.from_proto(computation_pb.literal.value)
    type_signature = computation_types.Type.from_proto(computation_pb.type)
    if not isinstance(type_signature, computation_types.TensorType):
      raise ValueError(
          'Expected `type_signature` to be a `federated_language.TensorType`,'
          f' found {type(type_signature)}.'
      )
    return Literal(value, type_signature)

  def to_proto(self) -> computation_pb2.Computation:
    """Returns a `computation_pb2.Computation` for this building block."""
    if self._proto is None:
      type_pb = self._type_signature.to_proto()
      value_pb = array.to_proto(
          self._value, dtype_hint=self.type_signature.dtype.type
      )
      literal_pb = computation_pb2.Literal(value=value_pb)
      self._proto = computation_pb2.Computation(
          type=type_pb,
          literal=literal_pb,
      )
    return self._proto

  @property
  def type_signature(self) -> computation_types.TensorType:
    return self._type_signature

  def children(self) -> Iterator[ComputationBuildingBlock]:
    return iter(())

  @property
  def value(self) -> object:
    return self._value

  def __eq__(self, other: object) -> bool:
    if self is other:
      return True
    elif not isinstance(other, Literal):
      return NotImplemented

    if self._type_signature != other._type_signature:
      return False
    if isinstance(self._value, np.ndarray) and isinstance(
        other._value, np.ndarray
    ):
      return np.array_equal(self._value, other._value)
    else:
      return self._value == other._value

  def __hash__(self):
    if self._hash is None:
      if isinstance(self._value, (np.generic, np.ndarray)):
        hashable_value = tuple(self._value.flatten().tolist())
      else:
        hashable_value = self._value
      self._hash = hash((hashable_value, self._type_signature))
    return self._hash

  def __repr__(self) -> str:
    if isinstance(self._value, np.ndarray):
      value_repr = (
          f'np.array({self._value.tolist()},'
          f' dtype=np.{self._value.dtype.type.__name__})'
      )
    else:
      value_repr = repr(self._value)
    return f'Literal({value_repr}, {self._type_signature!r})'


def _string_representation(
    comp: ComputationBuildingBlock,
    formatted: bool,
) -> str:
  """Returns the string representation of a `ComputationBuildingBlock`.

  This functions creates a `list` of strings representing the given `comp`;
  combines the strings in either a formatted or un-formatted representation; and
  returns the resulting string represetnation.

  Args:
    comp: An instance of a `ComputationBuildingBlock`.
    formatted: A boolean indicating if the returned string should be formatted.

  Raises:
    TypeError: If `comp` has an unepxected type.
  """

  def _join(components: Iterable[list[str]]) -> list[str]:
    """Returns a `list` of strings by combining each component in `components`.

    >>> _join([['a'], ['b'], ['c']])
    ['abc']

    >>> _join([['a', 'b', 'c'], ['d', 'e', 'f']])
    ['abcd', 'ef']

    This function is used to help track where new-lines should be inserted into
    the string representation if the lines are formatted.

    Args:
      components: A `list` where each element is a `list` of strings
        representing a part of the string of a `ComputationBuildingBlock`.
    """
    lines = ['']
    for component in components:
      lines[-1] = '{}{}'.format(lines[-1], component[0])
      lines.extend(component[1:])
    return lines

  def _indent(lines, indent_chars='  '):
    """Returns a `list` of strings indented across a slice."""
    return ['{}{}'.format(indent_chars, e) for e in lines]

  def _lines_for_named_comps(named_comps, formatted):
    """Returns a `list` of strings representing the given `named_comps`.

    Args:
      named_comps: A `list` of named computations, each being a pair consisting
        of a name (either a string, or `None`) and a `ComputationBuildingBlock`.
      formatted: A boolean indicating if the returned string should be
        formatted.
    """
    lines = []
    for index, (name, comp) in enumerate(named_comps):
      if index != 0:
        if formatted:
          lines.append([',', ''])
        else:
          lines.append([','])
      element_lines = _lines_for_comp(comp, formatted)
      if name is not None:
        element_lines = _join([
            ['{}='.format(name)],
            element_lines,
        ])
      lines.append(element_lines)
    return _join(lines)

  def _lines_for_comp(comp, formatted):
    """Returns a `list` of strings representing the given `comp`.

    Args:
      comp: An instance of a `ComputationBuildingBlock`.
      formatted: A boolean indicating if the returned string should be
        formatted.
    """
    if isinstance(comp, Block):
      lines = []
      variables_lines = _lines_for_named_comps(comp.locals, formatted)
      if formatted:
        variables_lines = _indent(variables_lines)
        lines.extend([['(let ', ''], variables_lines, ['', ' in ']])
      else:
        lines.extend([['(let '], variables_lines, [' in ']])
      result_lines = _lines_for_comp(comp.result, formatted)
      lines.append(result_lines)
      lines.append([')'])
      return _join(lines)
    elif isinstance(comp, Reference):
      if comp.context is not None:
        return ['{}@{}'.format(comp.name, comp.context)]
      else:
        return [comp.name]
    elif isinstance(comp, Selection):
      source_lines = _lines_for_comp(comp.source, formatted)
      if comp.name is not None:
        return _join([source_lines, ['.{}'.format(comp.name)]])
      else:
        return _join([source_lines, ['[{}]'.format(comp.index)]])
    elif isinstance(comp, Call):
      function_lines = _lines_for_comp(comp.function, formatted)
      if comp.argument is not None:
        argument_lines = _lines_for_comp(comp.argument, formatted)
        return _join([function_lines, ['('], argument_lines, [')']])
      else:
        return _join([function_lines, ['()']])
    elif isinstance(comp, CompiledComputation):
      return ['comp#{}'.format(comp.name)]
    elif isinstance(comp, Data):
      return [str(id(comp.content))]
    elif isinstance(comp, Intrinsic):
      return [comp.uri]
    elif isinstance(comp, Lambda):
      result_lines = _lines_for_comp(comp.result, formatted)
      if comp.parameter_type is None:
        param_name = ''
      else:
        param_name = comp.parameter_name
      lines = [['({} -> '.format(param_name)], result_lines, [')']]
      return _join(lines)
    elif isinstance(comp, Placement):
      placement_literal = placements.uri_to_placement_literal(comp.uri)
      return [placement_literal.name]
    elif isinstance(comp, Literal):
      return [str(comp.value)]
    elif isinstance(comp, Struct):
      if not comp:
        return ['<>']
      elements = comp.items()
      elements_lines = _lines_for_named_comps(elements, formatted)
      if formatted:
        elements_lines = _indent(elements_lines)
        lines = [['<', ''], elements_lines, ['', '>']]
      else:
        lines = [['<'], elements_lines, ['>']]
      return _join(lines)
    else:
      raise NotImplementedError(
          f'Unexpected building block found: {type(comp)}.'
      )

  lines = _lines_for_comp(comp, formatted)
  lines = [line.rstrip() for line in lines]
  if formatted:
    return '\n'.join(lines)
  else:
    return ''.join(lines)


def _structural_representation(comp: ComputationBuildingBlock):
  """Returns the structural string representation of the given `comp`.

  This functions creates and returns a string representing the structure of the
  abstract syntax tree for the given `comp`.

  Args:
    comp: An instance of a `ComputationBuildingBlock`.

  Raises:
    TypeError: If `comp` has an unepxected type.
  """
  padding_char = ' '

  def _get_leading_padding(string):
    """Returns the length of the leading padding for the given `string`."""
    for index, character in enumerate(string):
      if character != padding_char:
        return index
    return len(string)

  def _get_trailing_padding(string):
    """Returns the length of the trailing padding for the given `string`."""
    for index, character in enumerate(reversed(string)):
      if character != padding_char:
        return index
    return len(string)

  def _pad_left(lines, total_width):
    """Pads the beginning of each line in `lines` to the given `total_width`.

    >>>_pad_left(['aa', 'bb'], 4)
    ['  aa', '  bb',]

    Args:
      lines: A `list` of strings to pad.
      total_width: The length that each line in `lines` should be padded to.

    Returns:
      A `list` of lines with padding applied.
    """

    def _pad_line_left(line, total_width):
      current_width = len(line)
      assert current_width <= total_width
      padding = total_width - current_width
      return '{}{}'.format(padding_char * padding, line)

    return [_pad_line_left(line, total_width) for line in lines]

  def _pad_right(lines, total_width):
    """Pads the end of each line in `lines` to the given `total_width`.

    >>>_pad_right(['aa', 'bb'], 4)
    ['aa  ', 'bb  ']

    Args:
      lines: A `list` of strings to pad.
      total_width: The length that each line in `lines` should be padded to.

    Returns:
      A `list` of lines with padding applied.
    """

    def _pad_line_right(line, total_width):
      current_width = len(line)
      assert current_width <= total_width
      padding = total_width - current_width
      return '{}{}'.format(line, padding_char * padding)

    return [_pad_line_right(line, total_width) for line in lines]

  class Alignment(enum.Enum):
    LEFT = 1
    RIGHT = 2

  def _concatenate(lines_1, lines_2, align):
    """Concatenates two `list`s of strings.

    Concatenates two `list`s of strings by appending one list of strings to the
    other and then aligning lines of different widths by either padding the left
    or padding the right of each line to the width of the longest line.

    >>>_concatenate(['aa', 'bb'], ['ccc'], Alignment.LEFT)
    ['aa ', 'bb ', 'ccc']

    Args:
      lines_1: A `list` of strings.
      lines_2: A `list` of strings.
      align: An enum indicating how to align lines of different widths.

    Returns:
      A `list` of lines.
    """
    lines = lines_1 + lines_2
    longest_line = max(lines, key=len)
    longest_width = len(longest_line)
    if align is Alignment.LEFT:
      return _pad_right(lines, longest_width)
    elif align is Alignment.RIGHT:
      return _pad_left(lines, longest_width)

  def _calculate_inset_from_padding(
      left, right, preferred_padding, minimum_content_padding
  ):
    """Calculates the inset for the given padding.

    NOTE: This function is intended to only be called from `_fit_with_padding`.

    Args:
      left: A `list` of strings.
      right: A `list` of strings.
      preferred_padding: The preferred amount of non-negative padding between
        the lines in the fitted `list` of strings.
      minimum_content_padding: The minimum amount of non-negative padding
        allowed between the lines in the fitted `list` of strings.

    Returns:
      An integer.
    """
    assert preferred_padding >= 0
    assert minimum_content_padding >= 0

    trailing_padding = _get_trailing_padding(left[0])
    leading_padding = _get_leading_padding(right[0])
    inset = trailing_padding + leading_padding - preferred_padding
    for left_line, right_line in zip(left[1:], right[1:]):
      trailing_padding = _get_trailing_padding(left_line)
      leading_padding = _get_leading_padding(right_line)
      minimum_inset = (
          trailing_padding + leading_padding - minimum_content_padding
      )
      inset = min(inset, minimum_inset)
    return inset

  def _fit_with_inset(left, right, inset):
    r"""Concatenates the lines of two `list`s of strings.

    NOTE: This function is intended to only be called from `_fit_with_padding`.

    Args:
      left: A `list` of strings.
      right: A `list` of strings.
      inset: The amount of padding to remove or add when concatenating the
        lines.

    Returns:
      A `list` of lines.
    """
    lines = []
    for left_line, right_line in zip(left, right):
      if inset > 0:
        left_inset = 0
        trailing_padding = _get_trailing_padding(left_line)
        if trailing_padding > 0:
          left_inset = min(trailing_padding, inset)
          left_line = left_line[:-left_inset]
        if inset - left_inset > 0:
          leading_padding = _get_leading_padding(right_line)
          if leading_padding > 0:
            right_inset = min(leading_padding, inset - left_inset)
            right_line = right_line[right_inset:]
      padding = abs(inset) if inset < 0 else 0
      line = ''.join([left_line, padding_char * padding, right_line])
      lines.append(line)
    left_height = len(left)
    right_height = len(right)
    if left_height > right_height:
      lines.extend(left[right_height:])
    elif right_height > left_height:
      lines.extend(right[left_height:])
    longest_line = max(lines, key=len)
    longest_width = len(longest_line)
    shortest_line = min(lines, key=len)
    shortest_width = len(shortest_line)
    if shortest_width != longest_width:
      if left_height > right_height:
        lines = _pad_right(lines, longest_width)
      else:
        lines = _pad_left(lines, longest_width)
    return lines

  def _fit_with_padding(
      left, right, preferred_padding, minimum_content_padding=4
  ):
    r"""Concatenates the lines of two `list`s of strings.

    Concatenates the lines of two `list`s of strings by appending each line
    together using a padding. The same padding is used to append each line and
    the padding is calculated starting from the `preferred_padding` without
    going below `minimum_content_padding` on any of the lines. If the two
    `list`s of strings have different lengths, padding will be applied to
    maintain the length of each string in the resulting `list` of strings.

    >>>_fit_with_padding(['aa', 'bb'], ['ccc'])
    ['aa    cccc', 'bb        ']

    >>>_fit_with_padding(['aa          ', 'bb          '], ['          ccc'])
    ['aa    cccc', 'bb        ']

    Args:
      left: A `list` of strings.
      right: A `list` of strings.
      preferred_padding: The preferred amount of non-negative padding between
        the lines in the fitted `list` of strings.
      minimum_content_padding: The minimum amount of non-negative padding
        allowed between the lines in the fitted `list` of strings.

    Returns:
      A `list` of lines.
    """
    inset = _calculate_inset_from_padding(
        left, right, preferred_padding, minimum_content_padding
    )
    return _fit_with_inset(left, right, inset)

  def _get_node_label(comp):
    """Returns a string for node in the structure of the given `comp`."""
    if isinstance(comp, Block):
      return 'Block'
    elif isinstance(comp, Call):
      return 'Call'
    elif isinstance(comp, CompiledComputation):
      return 'Compiled({})'.format(comp.name)
    elif isinstance(comp, Data):
      return f'Data({id(comp.content)})'
    elif isinstance(comp, Intrinsic):
      return comp.uri
    elif isinstance(comp, Lambda):
      return 'Lambda({})'.format(comp.parameter_name)
    elif isinstance(comp, Reference):
      return 'Ref({})'.format(comp.name)
    elif isinstance(comp, Placement):
      return 'Placement'
    elif isinstance(comp, Selection):
      key = comp.name if comp.name is not None else comp.index
      return 'Sel({})'.format(key)
    elif isinstance(comp, Struct):
      return 'Struct'
    elif isinstance(comp, Literal):
      return f'Lit({comp.value})'
    else:
      raise TypeError('Unexpected type found: {}.'.format(type(comp)))

  def _lines_for_named_comps(named_comps):
    """Returns a `list` of strings representing the given `named_comps`.

    Args:
      named_comps: A `list` of named computations, each being a pair consisting
        of a name (either a string, or `None`) and a `ComputationBuildingBlock`.
    """
    lines = ['[']
    for index, (name, comp) in enumerate(named_comps):
      comp_lines = _lines_for_comp(comp)
      if name is not None:
        label = '{}='.format(name)
        comp_lines = _fit_with_padding([label], comp_lines, 0, 0)
      if index == 0:
        lines = _fit_with_padding(lines, comp_lines, 0, 0)
      else:
        lines = _fit_with_padding(lines, [','], 0, 0)
        lines = _fit_with_padding(lines, comp_lines, 1)
    lines = _fit_with_padding(lines, [']'], 0, 0)
    return lines

  def _lines_for_comp(comp):
    """Returns a `list` of strings representing the given `comp`.

    Args:
      comp: An instance of a `ComputationBuildingBlock`.
    """
    node_label = _get_node_label(comp)

    if isinstance(
        comp,
        (
            CompiledComputation,
            Data,
            Intrinsic,
            Placement,
            Reference,
            Literal,
        ),
    ):
      return [node_label]
    elif isinstance(comp, Block):
      variables_lines = _lines_for_named_comps(comp.locals)
      variables_width = len(variables_lines[0])
      variables_trailing_padding = _get_trailing_padding(variables_lines[0])
      leading_padding = variables_width - variables_trailing_padding
      edge_line = '{}/'.format(padding_char * leading_padding)
      variables_lines = _concatenate(
          [edge_line], variables_lines, Alignment.LEFT
      )

      result_lines = _lines_for_comp(comp.result)
      result_width = len(result_lines[0])
      leading_padding = _get_leading_padding(result_lines[0]) - 1
      trailing_padding = result_width - leading_padding - 1
      edge_line = '\\{}'.format(padding_char * trailing_padding)
      result_lines = _concatenate([edge_line], result_lines, Alignment.RIGHT)

      preferred_padding = len(node_label)
      lines = _fit_with_padding(
          variables_lines, result_lines, preferred_padding
      )
      leading_padding = _get_leading_padding(lines[0]) + 1
      node_line = '{}{}'.format(padding_char * leading_padding, node_label)
      return _concatenate([node_line], lines, Alignment.LEFT)
    elif isinstance(comp, Call):
      function_lines = _lines_for_comp(comp.function)
      function_width = len(function_lines[0])
      function_trailing_padding = _get_trailing_padding(function_lines[0])
      leading_padding = function_width - function_trailing_padding
      edge_line = '{}/'.format(padding_char * leading_padding)
      function_lines = _concatenate([edge_line], function_lines, Alignment.LEFT)

      if comp.argument is not None:
        argument_lines = _lines_for_comp(comp.argument)
        argument_width = len(argument_lines[0])
        leading_padding = _get_leading_padding(argument_lines[0]) - 1
        trailing_padding = argument_width - leading_padding - 1
        edge_line = '\\{}'.format(padding_char * trailing_padding)
        argument_lines = _concatenate(
            [edge_line], argument_lines, Alignment.RIGHT
        )

        preferred_padding = len(node_label)
        lines = _fit_with_padding(
            function_lines, argument_lines, preferred_padding
        )
      else:
        lines = function_lines
      leading_padding = _get_leading_padding(lines[0]) + 1
      node_line = '{}{}'.format(padding_char * leading_padding, node_label)
      return _concatenate([node_line], lines, Alignment.LEFT)
    elif isinstance(comp, Lambda):
      result_lines = _lines_for_comp(comp.result)
      leading_padding = _get_leading_padding(result_lines[0])
      node_line = '{}{}'.format(padding_char * leading_padding, node_label)
      edge_line = '{}|'.format(padding_char * leading_padding)
      return _concatenate([node_line, edge_line], result_lines, Alignment.LEFT)
    elif isinstance(comp, Selection):
      source_lines = _lines_for_comp(comp.source)
      leading_padding = _get_leading_padding(source_lines[0])
      node_line = '{}{}'.format(padding_char * leading_padding, node_label)
      edge_line = '{}|'.format(padding_char * leading_padding)
      return _concatenate([node_line, edge_line], source_lines, Alignment.LEFT)
    elif isinstance(comp, Struct):
      elements = comp.items()
      elements_lines = _lines_for_named_comps(elements)
      leading_padding = _get_leading_padding(elements_lines[0])
      node_line = '{}{}'.format(padding_char * leading_padding, node_label)
      edge_line = '{}|'.format(padding_char * leading_padding)
      return _concatenate(
          [node_line, edge_line], elements_lines, Alignment.LEFT
      )
    else:
      raise NotImplementedError(
          f'Unexpected building block found: {type(comp)}.'
      )

  lines = _lines_for_comp(comp)
  lines = [line.rstrip() for line in lines]
  return '\n'.join(lines)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines an abstract interface for objects that can be serialized."""

import abc


class Serializable(abc.ABC):

  @classmethod
  @abc.abstractmethod
  def from_bytes(cls, buffer: bytes) -> 'Serializable':
    """Deserializes the object from bytes."""
    raise NotImplementedError

  @abc.abstractmethod
  def to_bytes(self) -> bytes:
    """Serializes the object to bytes."""
    raise NotImplementedError



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# limitations under the License.
"""Utilities for working with dtypes."""

from collections.abc import Mapping
from typing import Union
import warnings

from federated_language.proto import data_type_pb2
import ml_dtypes
import numpy as np


# Mapping from `DataType` to `type[np.generic]`.
_PROTO_TO_DTYPE: Mapping[data_type_pb2.DataType, type[np.generic]] = {
    data_type_pb2.DataType.DT_BOOL: np.bool_,
    data_type_pb2.DataType.DT_INT8: np.int8,
    data_type_pb2.DataType.DT_INT16: np.int16,
    data_type_pb2.DataType.DT_INT32: np.int32,
    data_type_pb2.DataType.DT_INT64: np.int64,
    data_type_pb2.DataType.DT_UINT8: np.uint8,
    data_type_pb2.DataType.DT_UINT16: np.uint16,
    data_type_pb2.DataType.DT_UINT32: np.uint32,
    data_type_pb2.DataType.DT_UINT64: np.uint64,
    data_type_pb2.DataType.DT_HALF: np.float16,
    data_type_pb2.DataType.DT_FLOAT: np.float32,
    data_type_pb2.DataType.DT_DOUBLE: np.float64,
    data_type_pb2.DataType.DT_COMPLEX64: np.complex64,
    data_type_pb2.DataType.DT_COMPLEX128: np.complex128,
    data_type_pb2.DataType.DT_BFLOAT16: ml_dtypes.bfloat16,
    data_type_pb2.DataType.DT_STRING: np.str_,
}


def from_proto(
    dtype_pb: data_type_pb2.DataType,
) -> type[np.generic]:
  """Returns a `type[np.generic]` for the `dtype_pb`."""
  if dtype_pb in _PROTO_TO_DTYPE:
    return _PROTO_TO_DTYPE[dtype_pb]
  else:
    raise NotImplementedError(f'Unexpected dtype found: {dtype_pb}.')


# Mapping from `type[np.generic]` to `DataType`.
_DTYPE_TO_PROTO: Mapping[type[np.generic], data_type_pb2.DataType] = {
    np.bool_: data_type_pb2.DataType.DT_BOOL,
    np.int8: data_type_pb2.DataType.DT_INT8,
    np.int16: data_type_pb2.DataType.DT_INT16,
    np.int32: data_type_pb2.DataType.DT_INT32,
    np.int64: data_type_pb2.DataType.DT_INT64,
    np.uint8: data_type_pb2.DataType.DT_UINT8,
    np.uint16: data_type_pb2.DataType.DT_UINT16,
    np.uint32: data_type_pb2.DataType.DT_UINT32,
    np.uint64: data_type_pb2.DataType.DT_UINT64,
    np.float16: data_type_pb2.DataType.DT_HALF,
    np.float32: data_type_pb2.DataType.DT_FLOAT,
    np.float64: data_type_pb2.DataType.DT_DOUBLE,
    np.complex64: data_type_pb2.DataType.DT_COMPLEX64,
    np.complex128: data_type_pb2.DataType.DT_COMPLEX128,
    ml_dtypes.bfloat16: data_type_pb2.DataType.DT_BFLOAT16,
    np.str_: data_type_pb2.DataType.DT_STRING,
}


def to_proto(dtype: type[np.generic]) -> data_type_pb2.DataType:
  """Returns a `DataType` for the `dtype`."""
  if dtype in _DTYPE_TO_PROTO:
    return _DTYPE_TO_PROTO[dtype]
  else:
    raise NotImplementedError(f'Unexpected dtype found: {dtype}.')


def is_valid_dtype(dtype: type[np.generic]) -> bool:
  """Returns `True` if `dtype` is valid, otherwise `False`."""
  return dtype in _DTYPE_TO_PROTO


def can_cast(
    value: Union[bool, int, float, complex, str, bytes],
    dtype: type[np.generic],
) -> bool:
  """Returns `True` if `value` can be cast to `dtype`, otherwise `False`.

  This function is intended to be used to determine if the size of the `dtype`
  is capable of holding the `value`. This is useful, for example, when trying to
  infer the dtype of the `value`. This function is not intended to be used to
  determine if you **should** cast a the `value` to `dtype`.

  Args:
    value: The value to check.
    dtype: The dtype to check against.
  """

  # When encountering an overflow, numpy issues a `RuntimeWarning` for floating
  # dtypes and raises an `OverflowError` for integer dtypes.
  with warnings.catch_warnings():
    warnings.simplefilter(action='error', category=RuntimeWarning)
    try:
      np.asarray(value, dtype=dtype)
      return True
    except (OverflowError, RuntimeWarning):
      return False


def infer_dtype(
    obj: Union[bool, int, float, complex, str, bytes],
) -> type[np.generic]:
  """Returns a scalar numpy dtype for a Python scalar.

  Args:
    obj: A Python scalar.

  Returns:
    A scalar numpy dtype.
  """
  if isinstance(obj, bool):
    return np.bool_
  elif isinstance(obj, int):
    if can_cast(obj, np.int32):
      return np.int32
    elif can_cast(obj, np.int64):
      return np.int64
    else:
      raise ValueError(
          'Expected `obj` to be an `int` in the range'
          f' [{np.iinfo(np.int64).min}, {np.iinfo(np.int64).max}],'
          f' found: {obj}.'
      )
  elif isinstance(obj, float):
    return np.float32
  elif isinstance(obj, complex):
    return np.complex128
  elif isinstance(obj, (str, bytes)):
    return np.str_
  else:
    raise NotImplementedError(f'Unexpected obj found: {type(obj)}.')



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# limitations under the License.
"""A library of transformation functions for computation types."""

from collections.abc import Callable

from federated_language.types import computation_types


def strip_placement(
    type_signature: computation_types.Type,
) -> computation_types.Type:
  """Removes instances of `FederatedType` from `type_signature`."""

  def _remove_placement(type_signature):
    if isinstance(type_signature, computation_types.FederatedType):
      return type_signature.member, True
    return type_signature, False

  return transform_type_postorder(type_signature, _remove_placement)[0]


# TODO: b/134525440 - Unifying the recursive methods in type_analysis.
def transform_type_postorder(
    type_signature: computation_types.Type,
    transform_fn: Callable[
        [computation_types.Type], tuple[computation_types.Type, bool]
    ],
) -> tuple[computation_types.Type, bool]:
  """Walks type tree of `type_signature` postorder, calling `transform_fn`.

  Args:
    type_signature: Instance of `computation_types.Type` to transform
      recursively.
    transform_fn: Transformation function to apply to each node in the type tree
      of `type_signature`. Must be instance of Python function type.

  Returns:
    A possibly transformed version of `type_signature`, with each node in its
    tree the result of applying `transform_fn` to the corresponding node in
    `type_signature`.

  Raises:
    NotImplementedError: If the types don't match the specification above.
  """
  if isinstance(type_signature, computation_types.FederatedType):
    transformed_member, member_mutated = transform_type_postorder(
        type_signature.member, transform_fn
    )
    if member_mutated:
      type_signature = computation_types.FederatedType(
          transformed_member, type_signature.placement, type_signature.all_equal
      )
    type_signature, type_signature_mutated = transform_fn(type_signature)
    return type_signature, type_signature_mutated or member_mutated
  elif isinstance(type_signature, computation_types.SequenceType):
    transformed_element, element_mutated = transform_type_postorder(
        type_signature.element, transform_fn
    )
    if element_mutated:
      type_signature = computation_types.SequenceType(transformed_element)
    type_signature, type_signature_mutated = transform_fn(type_signature)
    return type_signature, type_signature_mutated or element_mutated
  elif isinstance(type_signature, computation_types.FunctionType):
    if type_signature.parameter is not None:
      transformed_parameter, parameter_mutated = transform_type_postorder(
          type_signature.parameter, transform_fn
      )
    else:
      transformed_parameter, parameter_mutated = (None, False)
    transformed_result, result_mutated = transform_type_postorder(
        type_signature.result, transform_fn
    )
    if parameter_mutated or result_mutated:
      type_signature = computation_types.FunctionType(
          transformed_parameter, transformed_result
      )
    type_signature, type_signature_mutated = transform_fn(type_signature)
    return type_signature, (
        type_signature_mutated or parameter_mutated or result_mutated
    )
  elif isinstance(type_signature, computation_types.StructType):
    elements = []
    elements_mutated = False
    for element in type_signature.items():
      transformed_element, element_mutated = transform_type_postorder(
          element[1], transform_fn
      )
      elements_mutated = elements_mutated or element_mutated
      elements.append((element[0], transformed_element))
    if elements_mutated:
      if isinstance(type_signature, computation_types.StructWithPythonType):
        type_signature = computation_types.StructWithPythonType(
            elements,
            type_signature.python_container,
        )
      else:
        type_signature = computation_types.StructType(elements)
    type_signature, type_signature_mutated = transform_fn(type_signature)
    return type_signature, type_signature_mutated or elements_mutated
  elif isinstance(
      type_signature,
      (
          computation_types.AbstractType,
          computation_types.PlacementType,
          computation_types.TensorType,
      ),
  ):
    return transform_fn(type_signature)
  else:
    raise NotImplementedError(f'Unexpected type found: {type(type_signature)}.')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for releasing values from a federated program."""

import abc
import asyncio
from collections.abc import Callable, Mapping, Sequence
import datetime
import typing
from typing import Generic, Optional, TypeVar, Union

import attrs
from federated_language.common_libs import py_typecheck
from federated_language.program import structure_utils
from federated_language.program import value_reference
import tree

# ReleaseManager's may release any value in addition to materializable values.
ReleasableValue = Union[
    object,
    value_reference.MaterializableValue,
]
ReleasableStructure = TypeVar(
    'ReleasableStructure',
    bound=structure_utils.Structure[ReleasableValue],
)
Key = TypeVar('Key')


class ReleasedValueNotFoundError(Exception):
  """Raised when a released value cannot be found."""

  def __init__(self, key: object):
    super().__init__(f'No released value found for key: {key}.')


class ReleaseManager(abc.ABC, Generic[ReleasableStructure, Key]):
  """An interface for releasing values from a federated program.

  A `federated_language.program.ReleaseManager` is used to release values from
  platform storage
  to customer storage in a federated program.
  """

  @abc.abstractmethod
  async def release(self, value: ReleasableStructure, key: Key) -> None:
    """Releases `value` from a federated program.

    An implementation of this interface should be specific about the types of
    `value` and `key` for this method and should document how the `key` will be
    used. This allows a federated program to understand how to create a `key`
    for the `value` before it is released. For example, a
    `federated_language.program.ReleaseManager` that releases metrics keyed by a
    strictly
    increasing integer might specify a `value` type of
    `Mapping[str, ReleasableValue]` and a `key` type of `int`.

    Args:
      value: A `federated_language.program.ReleasableStructure` to release.
      key: A value used to reference the released `value`.
    """
    raise NotImplementedError


class NotFilterableError(Exception):
  """Raised when the structure cannot be filtered."""


# Sentinel object used by the
# `federated_language.program.FilteringReleaseManager` to indicate that a
# subtree can be filtered when traversing structures of values and type
# signatures.
_FILTERED_SUBTREE = object()


class FilteringReleaseManager(ReleaseManager[ReleasableStructure, Key]):
  """A `federated_language.program.ReleaseManager` that filters values before releasing them.

  A `federated_language.program.FilteringReleaseManager` is a utility for
  filtering values
  before releasing the values and is used to release values from platform
  storage to customer storage in a federated program.

  Values are filtered using a `filter_fn` and released to the `release_manager`.

  The `filter_fn` is a `Callable` that has a single parameter `path` and returns
  a `bool`, and is used to filter values before they are released. A `path` is a
  tuple of indices and/or keys which uniquely identifies the position of the
  corresponding item in the `value`; `path` matches the expectations of the
  `tree` library.

  The `filter_fn` is applied to the items in the structure but not the structure
  itself. If all the items in a structure are filtered out, then the structure
  will be filtered out as well.

  For example:

  ```
  filtering_manager = federated_language.program.FilteringReleaseManager(
      release_manager=...,
      filter_fn=...,
  )

  value = {
    'loss': 1.0,
    'accuracy': 0.5,
  }
  await filtering_manager.release(value, ...)
  ```

  If `filter_fn` is:

  * `lambda _: True` then the entire structure is released.
  * `lambda _: False` then nothing is released.
  * `lambda path: path == ('loss',)` then `{'loss': 1.0}` is released.

  NOTE: The path `()` corresponds to the root of the structure; because the
  `filter_fn` is applied to the items in the structure but not the structure
  itself, this path can be used to filter individual values from structures of
  values.

  IMPORTANT: Most `federated_language.program.ReleasableStructure` can be
  filtered, including
  individual values, structures, and structures nested in `NamedTuple`s.
  However, the fields of a `NamedTuple` cannot be filtered.
  """

  def __init__(
      self,
      release_manager: ReleaseManager[ReleasableStructure, Key],
      filter_fn: Callable[[tuple[Union[str, int], ...]], bool],
  ):
    """Returns an initialized `federated_language.program.FilteringReleaseManager`.

    Args:
      release_manager: A `federated_language.program.ReleaseManager` used to
        release values to.
      filter_fn: A `Callable` used to filter values before they are released,
        this function has a single parameter `path` and returns a `bool`.
    """
    self._release_manager = release_manager
    self._filter_fn = filter_fn

  async def release(self, value: ReleasableStructure, key: Key) -> None:
    """Releases `value` from a federated program.

    Args:
      value: A `federated_language.program.ReleasableStructure` to release.
      key: A value used to reference the released `value`.

    Raises:
      NotFilterableError: If the `value` cannot be filtered.
    """

    def _filter_value(
        path: tuple[Union[str, int], ...],
        subtree: ReleasableStructure,
    ) -> Optional[Union[ReleasableStructure, type(_FILTERED_SUBTREE)]]:
      """The function to apply when filtering the `value`.

      This function is meant to be used with `tree.traverse_with_path` to filter
      the `value`. The function `tree.traverse_with_path` is used because the
      traversal functions from `tree` apply a function to the structure and
      the leaves, whereas the map functions only apply a function to the leaves.
      Additionally, `path` is used to determine which parts of the structure to
      filter.

      See https://tree.readthedocs.io/en/latest/api.html#tree.traverse for more
      information.

      Args:
        path: A tuple of indices and/or keys which uniquely identifies the
          position of `subtree` in the `value`.
        subtree: A substructure in `value`.

      Returns:
        A filtered value or `_FILTERED_SUBTREE` if the entire structure was
        filtered.

      Raises:
        NotFilterableError: If `subtree` cannot be filtered.
      """
      if tree.is_nested(subtree) and not attrs.has(type(subtree)):
        # TODO: b/224484886 - Downcasting to all handled types.
        subtree = typing.cast(
            Union[Sequence[object], Mapping[str, object]], subtree
        )
        if isinstance(subtree, Sequence):
          elements = [x for x in subtree if x is not _FILTERED_SUBTREE]
          if not elements:
            return _FILTERED_SUBTREE
          elif isinstance(subtree, py_typecheck.SupportsNamedTuple):
            if len(subtree) != len(elements):
              fields = list(type(subtree)._fields)
              missing_fields = [
                  k
                  for k, v in subtree._asdict().items()
                  if v is _FILTERED_SUBTREE
              ]
              raise NotFilterableError(
                  'The fields of a `NamedTuple` cannot be filtered. Expected '
                  f'{type(subtree)} to have fields {fields}, found it was '
                  f'missing fields {missing_fields}.'
              )

            return type(subtree)(*elements)
          else:
            # Assumes the `Sequence` has a constructor that accepts `elements`,
            #  this is safe because `tree` makes the same assumption.
            return type(subtree)(elements)  # pytype: disable=wrong-arg-count
        elif isinstance(subtree, Mapping):
          items = [
              (k, v) for k, v in subtree.items() if v is not _FILTERED_SUBTREE
          ]
          if not items:
            return _FILTERED_SUBTREE
          else:
            # Assumes the `Mapping` has a constructor that accepts `items`,
            # this is safe because `tree` makes the same assumption.
            return type(subtree)(items)  # pytype: disable=wrong-arg-count
        else:
          raise NotImplementedError(
              f'Unexpected subtree found: {type(subtree)}.'
          )
      else:
        if self._filter_fn(path):
          return None
        else:
          return _FILTERED_SUBTREE

    filtered_value = tree.traverse_with_path(
        _filter_value, value, top_down=False
    )

    if filtered_value is not _FILTERED_SUBTREE:
      await self._release_manager.release(filtered_value, key=key)


class GroupingReleaseManager(ReleaseManager[ReleasableStructure, Key]):
  """A `federated_language.program.ReleaseManager` that releases values to other release managers.

  A `federated_language.program.GroupingReleaseManager` is a utility for release
  values from a
  federated program to a collection of other release managers and is used to
  release values from platform storage to customer storage in a federated
  program.

  Values are released using each of the
  `federated_language.program.ReleaseManager`s in the
  given `release_managers`.
  """

  def __init__(
      self, release_managers: Sequence[ReleaseManager[ReleasableStructure, Key]]
  ):
    """Returns an initialized `federated_language.program.GroupingReleaseManager`.

    Args:
      release_managers: A sequence of
        `federated_language.program.ReleaseManager` used to release values to.

    Raises:
      ValueError: If `release_managers` is empty.
    """
    if not release_managers:
      raise ValueError('Expected `release_managers` to not be empty.')

    self._release_managers = release_managers

  async def release(self, value: ReleasableStructure, key: Key) -> None:
    """Releases `value` from a federated program.

    Args:
      value: A `federated_language.program.ReleasableStructure` to release.
      key: A value used to reference the released `value`.
    """
    await asyncio.gather(
        *[m.release(value, key=key) for m in self._release_managers]
    )


class PeriodicReleaseManager(ReleaseManager[ReleasableStructure, Key]):
  """A `federated_language.program.ReleaseManager` that releases values at regular intervals.

  A `federated_language.program.PeriodicReleaseManager` is a utility for
  releasing values at
  regular intervals and is used to release values from platform storage to
  customer storage in a federated program.

  The interval can be controlled at construction time by setting the
  `periodicity`. The `periodicity` can be a positive integer or
  `datetime.timedelta`. A `periodicity` of `3` means that every third value is
  released to the `release_manager`, and invoking `release` ten times will
  release the third, sixth, and ninth values. A `periodicity` of
  `datetime.timedelta(hours=3)` means that three hours after the previously
  released value the next value is released to the `release_manager`.

   NOTE: that a `periodicity` of one or a very small `datetime.timedelta` will
   release every value, making the
   `federated_language.program.PeriodicReleaseManager` a noop
   wrapper around the `release_manager`.
  """

  def __init__(
      self,
      release_manager: ReleaseManager[ReleasableStructure, Key],
      periodicity: Union[int, datetime.timedelta],
  ):
    """Returns an initialized `federated_language.program.PeriodicReleaseManager`.

    Args:
      release_manager: A `federated_language.program.ReleaseManager` used to
        release values to.
      periodicity: The interval to release values. Must be a positive integer or
        `datetime.timedelta`.

    Raises:
      ValueError: If `periodicity` is not a positive integer or
      `datetime.timedelta`.
    """
    if (isinstance(periodicity, int) and periodicity < 1) or (
        isinstance(periodicity, datetime.timedelta)
        and periodicity.total_seconds() < 1.0
    ):
      raise ValueError(
          'Expected `periodicity` to be a positive integer or'
          f' `datetime.timedelta`, found {periodicity}.'
      )

    self._release_manager = release_manager
    self._periodicity = periodicity
    if isinstance(periodicity, int):
      self._count = 0
    elif isinstance(periodicity, datetime.timedelta):
      self._timestamp = datetime.datetime.now()
    else:
      raise NotImplementedError(
          f'Unexpected periodicity found: {type(periodicity)}.'
      )

  async def release(self, value: ReleasableStructure, key: Key) -> None:
    """Releases `value` from a federated program.

    Args:
      value: A `federated_language.program.ReleasableStructure` to release.
      key: A value used to reference the released `value`.
    """
    if isinstance(self._periodicity, int):
      self._count += 1
      if self._count % self._periodicity == 0:
        await self._release_manager.release(value, key=key)
    elif isinstance(self._periodicity, datetime.timedelta):
      now = datetime.datetime.now()
      if now >= self._timestamp + self._periodicity:
        self._timestamp = now
        await self._release_manager.release(value, key=key)
    else:
      raise NotImplementedError(
          f'Unexpected periodicity found: {type(self._periodicity)}.'
      )


class DelayedReleaseManager(ReleaseManager[ReleasableStructure, Key]):
  """A `federated_language.program.ReleaseManager` that releases values after specified delay.

  A `federated_language.program.DelayedReleaseManager` is a utility for
  releasing values in a
  federated program, where releases only take place after a specified delay
  count. I.e., releases from platform storage to customer storage will take
  place only after a certain number of instances that the `.release()` method is
  called. After this delay, further calls to `.release()` will release values
  (in accordance with the `release_manager` that was provided).

  For example, in a federated program that runs for a long time, one may want to
  skip releasing values until the program has run for a sufficiently long-enough
  period.

  The delay count is specified at construction time by setting the `delay`
  argument (an integer). A `delay` of `3` means that all values will start to be
  released once `release` has been invoked at least three times.

   NOTE: that a `delay` of one will release every value, making the
   `federated_language.program.DelayedReleaseManager` a noop wrapper around the
   `release_manager`.
  """

  def __init__(
      self,
      release_manager: ReleaseManager[ReleasableStructure, Key],
      delay: int,
  ):
    """Returns an initialized `federated_language.program.DelayedReleaseManager`.

    Args:
      release_manager: A `federated_language.program.ReleaseManager` used to
        release values to.
      delay: The delay duration before releasing values. Must be a positive
        integer.

    Raises:
      ValueError: If `delay` is not positive.
    """
    if delay < 1:
      raise ValueError(f'The `delay` must be positive but found {delay}.')

    self._release_manager = release_manager
    self._count = 0
    self._delay = delay

  async def release(self, value: ReleasableStructure, key: Key) -> None:
    """Releases `value` from a federated program.

    Args:
      value: A `federated_language.program.ReleasableStructure` to release.
      key: A value used to reference the released `value`.
    """
    self._count += 1
    if self._count >= self._delay:
      await self._release_manager.release(value, key=key)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines the abstract interface for classes that represent computations."""

import abc

from federated_language.types import computation_types
from federated_language.types import typed_object


class Computation(typed_object.TypedObject, abc.ABC):
  """An abstract interface for all classes that represent computations."""

  @property
  @abc.abstractmethod
  def type_signature(self) -> computation_types.FunctionType:
    """Returns the type of this object."""
    raise NotImplementedError

  @abc.abstractmethod
  def __call__(self, *args, **kwargs):
    """Invokes the computation with the given arguments in the given context.

    Args:
      *args: The positional arguments.
      **kwargs: The keyword-based arguments.

    Returns:
      The result of invoking the computation, the exact form of which depends
      on the context.
    """
    raise NotImplementedError

  @abc.abstractmethod
  def __hash__(self) -> int:
    """Hashes the computation.

    Backends reserve the right to compile instances of
    `federated_language.Computation`, as they may need different
    representations or data structures altogether. As these backends need to be
    able to cache the result of compilation, we require that
    `federated_language.Computation` subclasses be hashable.

    Returns:
      Integer representing the hash value of the
      `federated_language.Computation`.
    """
    raise NotImplementedError



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# limitations under the License.
"""Utilities for type conversion, type checking, type inference, etc."""

import collections
from collections.abc import Hashable, Mapping, Sequence
import functools
import typing
from typing import Optional, Union

import attrs
from federated_language.common_libs import py_typecheck
from federated_language.common_libs import structure
from federated_language.types import computation_types
from federated_language.types import dtype_utils
from federated_language.types import typed_object
import numpy as np
import tree


def infer_type(arg: object) -> Optional[computation_types.Type]:
  """Infers the type of the argument (a `computation_types.Type` instance).

  WARNING: This function is only partially implemented.

  The kinds of arguments that are currently correctly recognized:
  * tensors, variables, and data sets
  * things that are convertible to tensors (including `numpy` arrays, builtin
    types, as well as `list`s and `tuple`s of any of the above, etc.)
  * nested lists, `tuple`s, `namedtuple`s, anonymous `tuple`s, `dict`,
    `OrderedDict`s, `attrs` classes, and `federated_language.TypedObject`s

  Args:
    arg: The argument, the type of which to infer.

  Returns:
    Either an instance of `computation_types.Type`, or `None` if the argument is
    `None`.
  """

  # TODO: b/224484886 - Downcasting to all handled types.
  arg = typing.cast(
      Union[
          None,
          typed_object.TypedObject,
          py_typecheck.SupportsNamedTuple,
          Mapping[Hashable, object],
          tuple[object, ...],
          list[object],
      ],
      arg,
  )
  if arg is None:
    return None
  elif isinstance(arg, typed_object.TypedObject):
    return arg.type_signature
  elif attrs.has(type(arg)):
    items = attrs.asdict(arg, recurse=False).items()
    return computation_types.StructWithPythonType(
        [(k, infer_type(v)) for k, v in items], type(arg)
    )
  elif isinstance(arg, py_typecheck.SupportsNamedTuple):
    items = arg._asdict().items()
    return computation_types.StructWithPythonType(
        [(k, infer_type(v)) for k, v in items], type(arg)
    )
  elif isinstance(arg, Mapping):
    items = arg.items()
    return computation_types.StructWithPythonType(
        [(k, infer_type(v)) for k, v in items], type(arg)
    )
  elif isinstance(arg, (tuple, list)):
    elements = []
    all_elements_named = True
    for element in arg:
      all_elements_named &= py_typecheck.is_name_value_pair(
          element, name_type=str
      )
      elements.append(infer_type(element))
    # If this is a tuple of (name, value) pairs, the caller most likely intended
    # this to be a StructType, so we avoid storing the Python container.
    if elements and all_elements_named:
      return computation_types.StructType(elements)
    else:
      return computation_types.StructWithPythonType(elements, type(arg))
  elif isinstance(arg, (np.generic, np.ndarray)):
    return computation_types.TensorType(arg.dtype, arg.shape)
  elif isinstance(arg, (bool, int, float, complex, str, bytes)):
    dtype = dtype_utils.infer_dtype(arg)
    return computation_types.TensorType(dtype)
  else:
    raise NotImplementedError(f'Unexpected arg found: {type(arg)}.')


def to_structure_with_type(
    obj: object, type_spec: computation_types.Type
) -> object:
  """Converts the containers in `obj` to those defined by `type_spec`.

  NOTE: This function does not convert any leaf in the structure.

  For example:

  >>> obj = 1
  >>> type_spec = federated_language.TensorType(np.int32)
  >>> federated_language.types.to_structure_with_type(obj, type_spec)
  1

  >>> obj = [1, 2, 3]
  >>> type_spec = federated_language.StructType([np.int32] * 3)
  >>> federated_language.types.to_structure_with_type(obj, type_spec)
  [1, 2, 3]

  >>> obj = [1, 2, 3]
  >>> type_spec = federated_language.StructType([
  >>>     ('a', np.int32),
  >>>     ('b', np.int32),
  >>>     ('c', np.int32),
  >>> ])
  >>> federated_language.types.to_structure_with_type(obj, type_spec)
  {'a': 1, 'b': 2, 'c': 3}

  Args:
    obj: A Python value.
    type_spec: The `federated_language.Type` to use convert `obj`.

  Returns:
    A Python value equivalent to `obj` with a structure matching `type_spec`.

  Raises:
    ValueError: If `obj` and `type_spec` do not match or a container does not
      have either all named or unnamed elements.
  """

  def _get_item(
      type_spec: computation_types.Type, key: Union[str, int]
  ) -> Union[computation_types.FederatedType, computation_types.StructType]:
    if isinstance(type_spec, computation_types.FederatedType):
      if not type_spec.all_equal:
        return type_spec.member
      else:
        type_spec = type_spec.member
    elif isinstance(type_spec, computation_types.SequenceType):
      return type_spec.element

    if isinstance(type_spec, computation_types.StructType):
      return type_spec[key]
    else:
      raise ValueError(
          'Expected `type_spec` to be a `federated_language.StructType`, found'
          f' {type(type_spec)}.'
      )

  def _get_container_cls(
      type_spec: computation_types.StructType,
  ) -> type[object]:
    container_cls = type_spec.python_container
    if container_cls is None:
      has_names = [name is not None for name, _ in type_spec.items()]
      if any(has_names):
        if not all(has_names):
          raise ValueError(
              'Expected `type_spec` to have either all named or unnamed'
              f' elements, found {type_spec}.'
          )
        container_cls = dict
      else:
        container_cls = list
    return container_cls

  def _get_container_spec(
      type_spec: computation_types.Type,
  ) -> tuple[type[object], Optional[list[Optional[str]]]]:
    if isinstance(type_spec, computation_types.FederatedType):
      if not type_spec.all_equal:
        return list, None
      else:
        type_spec = type_spec.member
    elif isinstance(type_spec, computation_types.SequenceType):
      return list, None

    if isinstance(type_spec, computation_types.StructType):
      names = [name for name, _ in type_spec.items()]
      container_cls = _get_container_cls(type_spec)
      return container_cls, names
    else:
      raise ValueError(
          'Expected `type_spec` to be a `federated_language.StructType`,'
          ' found {type(type_spec)}.'
      )

  def _to_structure(path: tuple[Union[str, int], ...], obj: object) -> object:

    if tree.is_nested(obj):
      container_type = functools.reduce(_get_item, path, type_spec)
      container_cls, names = _get_container_spec(container_type)

      if isinstance(obj, py_typecheck.SupportsNamedTuple):
        elements = obj._asdict().values()
      elif isinstance(obj, Mapping):
        elements = obj.values()
      elif isinstance(obj, Sequence):
        elements = obj
      else:
        raise ValueError(
            'Expected `obj` to be a `NamedTuple`, `Mapping`, or `Sequence`,'
            f' found {type(obj)}.'
        )

      if isinstance(container_cls, py_typecheck.SupportsNamedTuple):
        return container_cls(**dict(zip(names, elements)))
      elif issubclass(container_cls, Mapping):
        return container_cls(zip(names, elements))  # pylint: disable=too-many-function-args
      elif issubclass(container_cls, Sequence):
        return container_cls(elements)  # pylint: disable=too-many-function-args
      else:
        raise ValueError(
            'Expected `container_cls` to be a `NamedTuple`, `Mapping`, or'
            f' `Sequence`, found {container_cls}.'
        )
    else:
      return None

  if tree.is_nested(obj):
    return tree.traverse_with_path(_to_structure, obj, top_down=False)
  else:
    return obj


def _is_container_type_without_names(container_type: type[object]) -> bool:
  """Returns whether `container_type`'s elements are unnamed."""
  return issubclass(container_type, (list, tuple)) and not isinstance(
      container_type, py_typecheck.SupportsNamedTuple
  )


def _is_container_type_with_names(container_type: type[object]) -> bool:
  """Returns whether `container_type`'s elements are named."""
  return (
      isinstance(container_type, py_typecheck.SupportsNamedTuple)
      or attrs.has(container_type)
      or issubclass(container_type, dict)
  )


def type_to_py_container(value, type_spec: computation_types.Type):
  """Recursively convert `structure.Struct`s to Python containers.

  This is in some sense the inverse operation to
  `structure.from_container`.

  This function assumes some unique behavior with regards to
  `federated_language.SequenceType`.
  If the `value` is a list, it may yield other `federated_language.StructTypes`
  as well as
  Python types. Otherwise, it may only yield Python types.

  Args:
    value: A structure of anonymous tuples of values corresponding to
      `type_spec`.
    type_spec: The `federated_language.Type` to which value should conform,
      possibly including `computation_types.StructWithPythonType`.

  Returns:
    The input value, with containers converted to appropriate Python
    containers as specified by the `type_spec`.

  Raises:
    ValueError: If the conversion is not possible due to a mix of named
      and unnamed values, or if `value` contains names that are mismatched or
      not present in the corresponding index of `type_spec`.
  """
  if isinstance(type_spec, computation_types.FederatedType):
    if type_spec.all_equal:
      structure_type_spec = type_spec.member
    else:
      if not isinstance(value, list):
        raise TypeError(
            'Unexpected Python type for non-all-equal type '
            f'{type_spec}: expected `list`, found `{type(value)}`.'
        )
      return [
          type_to_py_container(element, type_spec.member) for element in value
      ]
  else:
    structure_type_spec = type_spec

  if isinstance(structure_type_spec, computation_types.SequenceType):
    element_type = structure_type_spec.element
    if isinstance(value, list):
      return [type_to_py_container(element, element_type) for element in value]
    else:
      # Assume that the type of value does not understand `Struct` and that the
      # value must yield Python containers.
      return value

  if not isinstance(structure_type_spec, computation_types.StructType):
    return value

  if not isinstance(value, structure.Struct):
    # NOTE: When encountering non-`structure.Struct`s, we assume that
    # this means that we're attempting to re-convert a value that
    # already has the proper containers, and we short-circuit to
    # avoid re-converting. This is a possibly dangerous assumption.
    return value

  container_type = structure_type_spec.python_container

  # Ensure that names are only added, not mismatched or removed
  names_from_value = structure.name_list_with_nones(value)
  names_from_type_spec = structure.name_list_with_nones(structure_type_spec)
  for value_name, type_name in zip(names_from_value, names_from_type_spec):
    if value_name is not None:
      if value_name != type_name:
        raise ValueError(
            f'Cannot convert value with field name `{value_name}` into a '
            f'type with field name `{type_name}`.'
        )

  num_named_elements = len(dir(structure_type_spec))
  num_unnamed_elements = len(structure_type_spec) - num_named_elements
  if num_named_elements > 0 and num_unnamed_elements > 0:
    raise ValueError(
        f'Cannot represent value {value} with a Python container because it'
        ' contains a mix of named and unnamed elements.\n\nNOTE: this was'
        ' previously allowed when using the'
        ' `federated_language.structure.Struct` container. This support has'
        ' been removed: please change to use structures with either all-named'
        ' or all-unnamed fields.'
    )
  if container_type is None:
    if num_named_elements:
      container_type = collections.OrderedDict
    else:
      container_type = tuple

  # Avoid projecting the `structure.StructType`d value into a Python
  # container that is not supported.
  if num_named_elements > 0 and _is_container_type_without_names(
      container_type
  ):
    raise ValueError(
        'Cannot represent value {} with named elements '
        'using container type {} which does not support names. In the '
        'typesystem, this corresponds to an implicit downcast'.format(
            value, container_type
        )
    )
  if _is_container_type_with_names(container_type) and len(
      dir(structure_type_spec)
  ) != len(value):
    # If the type specifies the names, we have all the information we need.
    # Otherwise we must raise here.
    raise ValueError(
        'When packaging as a Python value which requires names, '
        'the type spec must have all names specified. Found '
        '{} names in type spec {} of length {}, with requested'
        'python type {}.'.format(
            len(dir(structure_type_spec)),
            structure_type_spec,
            len(value),
            container_type,
        )
    )

  elements = []
  for index, (elem_name, elem_type) in enumerate(structure_type_spec.items()):
    element = type_to_py_container(value[index], elem_type)

    if elem_name is None:
      elements.append(element)
    else:
      elements.append((elem_name, element))

  if isinstance(container_type, py_typecheck.SupportsNamedTuple) or attrs.has(
      container_type
  ):
    # The namedtuple and attr.s class constructors cannot interpret a list of
    # (name, value) tuples; instead call constructor using kwargs. Note that
    # these classes already define an order of names internally, so order does
    # not matter.
    return container_type(**dict(elements))
  else:
    # E.g., tuple and list when elements only has values, but also `dict`,
    # `collections.OrderedDict`, or `structure.Struct` when
    # elements has (name, value) tuples.
    return container_type(elements)  # pytype: disable=wrong-arg-count



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A factory of intrinsics for use in composing federated computations."""

from typing import NoReturn
import warnings

from federated_language.common_libs import structure
from federated_language.compiler import building_block_factory
from federated_language.compiler import building_blocks
from federated_language.compiler import intrinsic_defs
from federated_language.context_stack import context
from federated_language.context_stack import context_stack_impl
from federated_language.context_stack import symbol_binding_context
from federated_language.federated_context import value_impl
from federated_language.federated_context import value_utils
from federated_language.types import array_shape
from federated_language.types import computation_types
from federated_language.types import placements
from federated_language.types import type_analysis
from federated_language.types import type_factory
import numpy as np


def _bind_comp_as_reference(comp):
  fc_context = context_stack_impl.context_stack.current
  if not isinstance(fc_context, symbol_binding_context.SymbolBindingContext):
    raise context.ContextError(
        f'Attempted to construct an intrinsic in context {fc_context} which '
        ' does not support binding references.'
    )
  return fc_context.bind_computation_to_reference(comp)


def federated_aggregate(
    value, zero, accumulate, merge, report
) -> value_impl.Value:
  """Aggregates `value` from `federated_language.CLIENTS` to `federated_language.SERVER`.

  This generalized aggregation function admits multi-layered architectures that
  involve one or more intermediate stages to handle scalable aggregation across
  a very large number of participants.

  The multi-stage aggregation process is defined as follows:

  * Clients are organized into groups. Within each group, a set of all the
    member constituents of `value` contributed by clients in the group are first
    reduced using reduction operator `accumulate` with `zero` as the zero in the
    algebra. If members of `value` are of type `T`, and `zero` (the result of
    reducing an empty set) is of type `U`, the reduction operator `accumulate`
    used at this stage should be of type `(<U,T> -> U)`. The result of this
    stage is a set of items of type `U`, one item for each group of clients.

  * Next, the `U`-typed items generated by the preceding stage are merged using
    the binary commutative associative operator `merge` of type `(<U,U> -> U)`.
    The result of this stage is a single top-level `U` that emerges at the root
    of the hierarchy at the `federated_language.SERVER`. Actual implementations
    may structure
    this step as a cascade of multiple layers.

  * Finally, the `U`-typed result of the reduction performed in the preceding
    stage is projected into the result value using `report` as the mapping
    function (for example, if the structures being merged consist of counters,
    this final step might include computing their ratios).

  Args:
    value: A value of a federated type placed at `federated_language.CLIENTS` to
      aggregate.
    zero: The zero of type `U` in the algebra of reduction operators, as
      described above.
    accumulate: The reduction operator to use in the first stage of the process.
      If `value` is of type `{T}@CLIENTS`, and `zero` is of type `U`, this
      operator should be of type `(<U,T> -> U)`.
    merge: The reduction operator to employ in the second stage of the process.
      Must be of type `(<U,U> -> U)`, where `U` is as defined above.
    report: The projection operator to use at the final stage of the process to
      compute the final result of aggregation. If the intended result to be
      returned by `federated_language.federated_aggregate` is of type
      `R@SERVER`, this operator must be of type `(U -> R)`.

  Returns:
    A representation on the `federated_language.SERVER` of the result of
    aggregating `value`
    using the multi-stage process described above.

  Raises:
    TypeError: If the arguments are not of the types specified above.
  """
  value = value_impl.to_value(value, type_spec=None)
  value = value_utils.ensure_federated_value(
      value, placements.CLIENTS, 'value to be aggregated'
  )

  zero = value_impl.to_value(zero, type_spec=None)

  accumulate = value_impl.to_value(
      accumulate,
      type_spec=None,
      parameter_type_hint=computation_types.StructType([
          zero.type_signature,
          value.type_signature.member,  # pytype: disable=attribute-error
      ]),
  )
  if not isinstance(accumulate.type_signature, computation_types.FunctionType):
    raise ValueError(
        'Expected `accumulate.type_signature` to be a `FunctionType`,'
        f' found {type(accumulate.type_signature)}.'
    )

  merge = value_impl.to_value(
      merge,
      type_spec=None,
      parameter_type_hint=computation_types.StructType(
          [accumulate.type_signature.result] * 2
      ),
  )
  if not isinstance(merge.type_signature, computation_types.FunctionType):
    raise ValueError(
        'Expected `merge.type_signature` to be a `FunctionType`,'
        f' found {type(merge.type_signature)}.'
    )

  report = value_impl.to_value(
      report,
      type_spec=None,
      parameter_type_hint=merge.type_signature.result,
  )
  if not isinstance(report.type_signature, computation_types.FunctionType):
    raise ValueError(
        'Expected `report.type_signature` to be a `FunctionType`,'
        f' found {type(report.type_signature)}.'
    )

  if not accumulate.type_signature.parameter[0].is_assignable_from(  # pytype: disable=attribute-error
      zero.type_signature
  ):
    raise TypeError(
        'Expected `zero` to be assignable to type'
        f' {accumulate.type_signature.parameter[0]}, but was of incompatible'  # pytype: disable=attribute-error
        f' type {zero.type_signature}.'
    )

  accumulate_type_expected = type_factory.reduction_op(
      accumulate.type_signature.result,
      value.type_signature.member,  # pytype: disable=attribute-error
  )
  merge_type_expected = type_factory.reduction_op(
      accumulate.type_signature.result, accumulate.type_signature.result
  )
  report_type_expected = computation_types.FunctionType(
      merge.type_signature.result, report.type_signature.result
  )
  for op_name, op, type_expected in [
      ('accumulate', accumulate, accumulate_type_expected),
      ('merge', merge, merge_type_expected),
      ('report', report, report_type_expected),
  ]:
    if not type_expected.is_assignable_from(op.type_signature):
      raise TypeError(
          'Expected parameter `{}` to be of type {}, but received {} instead.'
          .format(op_name, type_expected, op.type_signature)
      )

  comp = building_block_factory.create_federated_aggregate(
      value.comp, zero.comp, accumulate.comp, merge.comp, report.comp
  )
  comp = _bind_comp_as_reference(comp)
  return value_impl.Value(comp)


def federated_broadcast(value):
  """Broadcasts a federated value from the `federated_language.SERVER` to the `federated_language.CLIENTS`.

  Args:
    value: A value of a federated type placed at the
      `federated_language.SERVER`, all members of which are equal (the
      `federated_language.FederatedType.all_equal` property of `value` is
      `True`).

  Returns:
    A representation of the result of broadcasting: a value of a federated type
    placed at the `federated_language.CLIENTS`, all members of which are
    equal.

  Raises:
    TypeError: If the argument is not a federated value placed at the
      `federated_language.SERVER`.
  """
  value = value_impl.to_value(value, type_spec=None)
  value = value_utils.ensure_federated_value(
      value, placements.SERVER, 'value to be broadcasted'
  )

  if not value.type_signature.all_equal:  # pytype: disable=attribute-error
    raise TypeError('The broadcasted value should be equal at all locations.')

  comp = building_block_factory.create_federated_broadcast(value.comp)
  comp = _bind_comp_as_reference(comp)
  return value_impl.Value(comp)


def federated_eval(fn, placement):
  """Evaluates a federated computation at `placement`, returning the result.

  Args:
    fn: A no-arg computation.
    placement: The desired result placement (either `federated_language.SERVER`
      or `federated_language.CLIENTS`).

  Returns:
    A federated value with the given placement `placement`.

  Raises:
    TypeError: If the arguments are not of the appropriate types.
  """
  # TODO: b/113112108 - Verify that neither the value, nor any of its parts
  # are of a federated type.

  fn = value_impl.to_value(fn, type_spec=None)

  if not isinstance(fn.type_signature, computation_types.FunctionType):
    raise ValueError(
        'Expected `fn.type_signature` to be a `FunctionType`,'
        f' found {type(fn.type_signature)}.'
    )
  if fn.type_signature.parameter is not None:
    raise ValueError(
        'Expected `fn.type_signature` to have no parameters, found'
        f' {fn.type_signature.parameter}.'
    )

  comp = building_block_factory.create_federated_eval(fn.comp, placement)
  comp = _bind_comp_as_reference(comp)
  return value_impl.Value(comp)


def federated_map(fn, arg):
  """Maps a federated value pointwise using a mapping function.

  The function `fn` is applied separately across the group of devices
  represented by the placement type of `arg`. For example, if `value` has
  placement type `federated_language.CLIENTS`, then `fn` is applied to each
  client
  individually. In particular, this operation does not alter the placement of
  the federated value.

  Args:
    fn: A mapping function to apply pointwise to member constituents of `arg`.
      The parameter of this function must be of the same type as the member
      constituents of `arg`.
    arg: A value of a federated type (or a value that can be implicitly
      converted into a federated type, e.g., by zipping) placed at
      `federated_language.CLIENTS` or `federated_language.SERVER`.

  Returns:
    A federated value with the same placement as `arg` that represents the
    result of `fn` on the member constituent of `arg`.

  Raises:
    TypeError: If the arguments are not of the appropriate types.
  """
  # TODO: b/113112108 - Possibly lift the restriction that the mapped value
  # must be placed at the server or clients. Would occur after adding support
  # for placement labels in the federated types, and expanding the type
  # specification of the intrinsic this is based on to work with federated
  # values of arbitrary placement.

  arg = value_impl.to_value(arg, type_spec=None)
  arg = value_utils.ensure_federated_value(arg, label='value to be mapped')

  fn = value_impl.to_value(
      fn, type_spec=None, parameter_type_hint=arg.type_signature.member  # pytype: disable=attribute-error
  )
  if not isinstance(fn.type_signature, computation_types.FunctionType):
    raise ValueError(
        'Expected `fn.type_signature` to be a `FunctionType`,'
        f' found {type(fn.type_signature)}.'
    )
  if fn.type_signature.parameter is None:
    raise ValueError(
        'Expected `fn.type_signature` to have a parameter, found none.'
    )

  if not fn.type_signature.parameter.is_assignable_from(
      arg.type_signature.member  # pytype: disable=attribute-error
  ):
    raise TypeError(
        'The mapping function expects a parameter of type {}, but member '
        'constituents of the mapped value are of incompatible type {}.'.format(
            fn.type_signature.parameter,
            arg.type_signature.member,  # pytype: disable=attribute-error
        )
    )

  # TODO: b/144384398 - Change structure to one that maps the placement type
  # to the building_block function that fits it, in a way that allows the
  # appropriate type checks.
  if arg.type_signature.placement is placements.SERVER:  # pytype: disable=attribute-error
    if not arg.type_signature.all_equal:  # pytype: disable=attribute-error
      raise TypeError(
          'Arguments placed at {} should be equal at all locations.'.format(
              placements.SERVER
          )
      )
    comp = building_block_factory.create_federated_apply(fn.comp, arg.comp)
  elif arg.type_signature.placement is placements.CLIENTS:  # pytype: disable=attribute-error
    comp = building_block_factory.create_federated_map(fn.comp, arg.comp)
  else:
    raise TypeError(
        'Expected `arg` to have a type with a supported placement, '
        'found {}.'.format(arg.type_signature.placement)  # pytype: disable=attribute-error
    )

  comp = _bind_comp_as_reference(comp)
  return value_impl.Value(comp)


def federated_map_all_equal(fn, arg):
  """`federated_map` with the `all_equal` bit set in the `arg` and return."""
  # TODO: b/113112108 - Possibly lift the restriction that the mapped value
  # must be placed at the clients after adding support for placement labels
  # in the federated types, and expanding the type specification of the
  # intrinsic this is based on to work with federated values of arbitrary
  # placement.
  arg = value_impl.to_value(arg, type_spec=None)
  arg = value_utils.ensure_federated_value(
      arg, placements.CLIENTS, 'value to be mapped'
  )

  fn = value_impl.to_value(
      fn, type_spec=None, parameter_type_hint=arg.type_signature.member  # pytype: disable=attribute-error
  )
  if not isinstance(fn.type_signature, computation_types.FunctionType):
    raise ValueError(
        'Expected `fn.type_signature` to be a `FunctionType`,'
        f' found {type(fn.type_signature)}.'
    )
  if fn.type_signature.parameter is None:
    raise ValueError(
        'Expected `fn.type_signature` to have a parameter, found none.'
    )

  if not fn.type_signature.parameter.is_assignable_from(
      arg.type_signature.member  # pytype: disable=attribute-error
  ):
    raise TypeError(
        'The mapping function expects a parameter of type {}, but member '
        'constituents of the mapped value are of incompatible type {}.'.format(
            fn.type_signature.parameter,
            arg.type_signature.member,  # pytype: disable=attribute-error
        )
    )

  comp = building_block_factory.create_federated_map_all_equal(
      fn.comp, arg.comp
  )
  comp = _bind_comp_as_reference(comp)
  return value_impl.Value(comp)


def federated_mean(value, weight=None):
  """Computes a `federated_language.SERVER` mean of `value` placed on `federated_language.CLIENTS`.

  For values `v_1, ..., v_k`, and weights `w_1, ..., w_k`, this means
  `sum_{i=1}^k (w_i * v_i) / sum_{i=1}^k w_i`.

  Args:
    value: The value of which the mean is to be computed. Must be of a federated
      type placed at `federated_language.CLIENTS`. The value may be structured,
      e.g., its member constituents can be named tuples. The tensor types that
      the value is composed of must be floating-point or complex.
    weight: An optional weight, a federated integer or floating-point tensor
      value, also placed at `federated_language.CLIENTS`.

  Returns:
    A representation at the `federated_language.SERVER` of the mean of the
    member constituents
    of `value`, optionally weighted with `weight` if specified (otherwise, the
    member constituents contributed by all clients are equally weighted).

  Raises:
    TypeError: If `value` is not a federated value placed at
    `federated_language.CLIENTS`,
      or if `weight` is not a federated integer or a floating-point tensor with
      the matching placement.
  """
  # TODO: b/120439632 - Possibly allow the weight to be either structured or
  # non-scalar, e.g., for the case of averaging a convolutional layer, when
  # we would want to use a different weight for every filter, and where it
  # might be cumbersome for users to have to manually slice and assemble a
  # variable.

  value = value_impl.to_value(value, type_spec=None)
  value = value_utils.ensure_federated_value(
      value, placements.CLIENTS, 'value to be averaged'
  )
  if not type_analysis.is_average_compatible(value.type_signature):
    raise TypeError(
        'The value type {} is not compatible with the average operator.'.format(
            value.type_signature
        )
    )

  if weight is not None:
    weight = value_impl.to_value(weight, type_spec=None)
    weight = value_utils.ensure_federated_value(
        weight, placements.CLIENTS, 'weight to use in averaging'
    )
    if (
        not isinstance(weight.type_signature, computation_types.FederatedType)
        or not isinstance(
            weight.type_signature.member, computation_types.TensorType
        )
        or not array_shape.is_shape_scalar(weight.type_signature.member.shape)
    ):
      raise TypeError(
          'The weight type {} is not a federated scalar.'.format(
              weight.type_signature
          )
      )
    if not np.issubdtype(
        weight.type_signature.member.dtype,  # pytype: disable=attribute-error
        np.integer,
    ) and not np.issubdtype(
        weight.type_signature.member.dtype,  # pytype: disable=attribute-error
        np.floating,
    ):
      raise TypeError(
          'The weight type {} is not a federated integer or floating-point '
          'tensor.'.format(weight.type_signature)
      )

  weight_comp = None if weight is None else weight.comp
  comp = building_block_factory.create_federated_mean(value.comp, weight_comp)
  comp = _bind_comp_as_reference(comp)
  return value_impl.Value(comp)


def federated_min(value: object) -> value_impl.Value:
  """Computes a min at `federated_language.SERVER` of a `value` placed on the `federated_language.CLIENTS`.

  Args:
    value: A value of a federated type placed at the
      `federated_language.CLIENTS`.

  Returns:
    A representation of the min of the member constituents of `value` placed on
    the `federated_language.SERVER`.

  Raises:
    ValueError: If the argument is not a federated value placed at
      `federated_language.CLIENTS` compatible with min.
  """
  value = value_impl.to_value(value, type_spec=None)
  value = value_utils.ensure_federated_value(
      value, placements.CLIENTS, 'value to take min of'
  )
  if not type_analysis.is_min_max_compatible(value.type_signature):
    raise ValueError(
        'The value type {} is not compatible with the min operator.'.format(
            value.type_signature
        )
    )
  comp = building_block_factory.create_federated_min(value.comp)
  comp = _bind_comp_as_reference(comp)
  return value_impl.Value(comp)


def federated_max(value: object) -> value_impl.Value:
  """Computes a max at `federated_language.SERVER` of a `value` placed on the `federated_language.CLIENTS`.

  Args:
    value: A value of a federated type placed at the
      `federated_language.CLIENTS`.

  Returns:
    A representation of the max of the member constituents of `value` placed on
    the `federated_language.SERVER`.

  Raises:
    ValueError: If the argument is not a federated value placed at
      `federated_language.CLIENTS` compatible with max.
  """
  value = value_impl.to_value(value, type_spec=None)
  value = value_utils.ensure_federated_value(
      value, placements.CLIENTS, 'value to take max of'
  )
  if not type_analysis.is_min_max_compatible(value.type_signature):
    raise ValueError(
        'The value type {} is not compatible with the max operator.'.format(
            value.type_signature
        )
    )
  comp = building_block_factory.create_federated_max(value.comp)
  comp = _bind_comp_as_reference(comp)
  return value_impl.Value(comp)


def federated_sum(value):
  """Computes a sum at `federated_language.SERVER` of a `value` placed on the `federated_language.CLIENTS`.

  To sum integer values with stronger privacy properties, consider using
  `federated_language.federated_secure_sum_bitwidth`.

  Args:
    value: A value of a federated type placed at the
      `federated_language.CLIENTS`.

  Returns:
    A representation of the sum of the member constituents of `value` placed
    on the `federated_language.SERVER`.

  Raises:
    TypeError: If the argument is not a federated value placed at
      `federated_language.CLIENTS`.
  """
  value = value_impl.to_value(value, type_spec=None)
  value = value_utils.ensure_federated_value(
      value, placements.CLIENTS, 'value to be summed'
  )
  type_analysis.check_is_sum_compatible(value.type_signature)
  comp = building_block_factory.create_federated_sum(value.comp)
  comp = _bind_comp_as_reference(comp)
  return value_impl.Value(comp)


def federated_value(value, placement):
  """Returns a federated value at `placement`, with `value` as the constituent.

  Deprecation warning: Using `federated_language.federated_value` with arguments
  other than
  simple Python constants is deprecated. When placing the result of a
  `federated_language.tensorflow.computation`, prefer
  `federated_language.federated_eval`.

  Args:
    value: A value of a non-federated type to be placed.
    placement: The desired result placement (either `federated_language.SERVER`
      or `federated_language.CLIENTS`).

  Returns:
    A federated value with the given placement `placement`, and the member
    constituent `value` equal at all locations.

  Raises:
    TypeError: If the arguments are not of the appropriate types.
  """
  if isinstance(value, value_impl.Value):
    warnings.warn(
        (
            'Deprecation warning: Using `federated_language.federated_value`'
            ' with arguments other than simple Python constants is deprecated.'
            ' When placing the result of a'
            ' `federated_language.tensorflow.computation`, prefer'
            ' `federated_language.federated_eval`.'
        ),
        DeprecationWarning,
    )
  value = value_impl.to_value(value, type_spec=None)
  if type_analysis.contains(
      value.type_signature,
      lambda t: isinstance(t, computation_types.FederatedType),
  ):
    raise TypeError(
        'Cannot place value {} containing federated types at '
        'another placement; requested to be placed at {}.'.format(
            value, placement
        )
    )

  comp = building_block_factory.create_federated_value(value.comp, placement)
  comp = _bind_comp_as_reference(comp)
  return value_impl.Value(comp)


def federated_zip(value):
  """Converts an N-tuple of federated values into a federated N-tuple value.

  Args:
    value: A value of a named tuple type, the elements of which are federated
      values with the same placement.

  Returns:
    A federated value placed at the same location as the members of `value`, in
    which every member component is a named tuple that consists of the
    corresponding member components of the elements of `value`.

  Raises:
    TypeError: If the argument is not a named tuple of federated values with the
      same placement.
  """
  # TODO: b/113112108 - We use the iterate/unwrap approach below because
  # our type system is not powerful enough to express the concept of
  # "an operation that takes tuples of T of arbitrary length", and therefore
  # the intrinsic federated_zip must only take a fixed number of arguments,
  # here fixed at 2. There are other potential approaches to getting around
  # this problem (e.g. having the operator act on sequences and thereby
  # sidestepping the issue) which we may want to explore.
  value = value_impl.to_value(value, type_spec=None)
  if not isinstance(value.type_signature, computation_types.StructType):
    raise ValueError(
        'Expected `value.type_signature` to be a `StructType`,'
        f' found {type(value.type_signature)}.'
    )

  comp = building_block_factory.create_federated_zip(value.comp)
  comp = _bind_comp_as_reference(comp)
  return value_impl.Value(comp)


def _select_parameter_mismatch(
    param_type,
    type_desc,
    name,
    secure,
    expected_type=None,
) -> NoReturn:
  """Throws a `TypeError` indicating a mismatched `select` parameter type."""
  secure_string = '_secure' if secure else ''
  intrinsic_name = f'federated{secure_string}_select'
  message = f'Expected `{intrinsic_name}` parameter `{name}` to be {type_desc}'
  if expected_type is None:
    raise TypeError(f'{message}, found value of type {param_type}')
  else:
    raise TypeError(
        f'{message}, {param_type} is not assignable to {expected_type}.'
    )


def _check_select_keys_type(
    keys_type: computation_types.Type, secure: bool
) -> None:
  """Checks the federated select keys types."""
  if (
      not isinstance(keys_type, computation_types.FederatedType)
      or keys_type.placement is not placements.CLIENTS
  ):
    _select_parameter_mismatch(
        keys_type, 'a federated value placed at clients', 'client_keys', secure
    )
  if not (
      isinstance(keys_type.member, computation_types.TensorType)
      and keys_type.member.dtype == np.int32
      and len(keys_type.member.shape) == 1
      and keys_type.member.shape[0] is not None
  ):
    _select_parameter_mismatch(
        keys_type.member,  # pytype: disable=attribute-error
        'a rank-1 tensor with statically known shape and `np.int32` dtype',
        'client_keys.type_signature.member',
        secure,
    )


def federated_select(client_keys, max_key, server_val, select_fn):
  """Sends selected values from a server database to clients.

  Args:
    client_keys: `federated_language.CLIENTS`-placed one-dimensional fixed-size
      non-negative `int32` keys used to select values from `database` to load
      for each client.
    max_key: A `federated_language.SERVER`-placed `int32` indicating the maximum
      value of any of `client_keys` (so that all client keys are in the range
      `[0, max_key]`, inclusive). Lower values may permit more optimizations.
    server_val: `federated_language.SERVER`-placed value used as an input to
      `select_fn`.
    select_fn: A `federated_language.Computation` which accepts unplaced
      `server_val` and a `int32` client key and returns a value to be sent to
      the client. `select_fn` should be deterministic (nonrandom).

  Returns:
    `federated_language.CLIENTS`-placed sequences of values returned from
    `select_fn`. In each
    sequence, the order of values will match the order of keys in the
    corresponding `client_keys` tensor. For example, a client with keys
    `[1, 2, ...]` will receive a sequence of values
    `[select_fn(server_val, 1), select_fn(server_val, 2), ...]`.

  Raises:
    TypeError: If `client_keys` is not of type `{int32[N]}@CLIENTS`, if
      `max_key` is not of type `int32@SERVER`, if `server_val` is not a
      server-placed value (`S@SERVER`), or if `select_fn` is not a function
      of type `<S, int32> -> RESULT`.
  """
  return _federated_select(
      client_keys, max_key, server_val, select_fn, secure=False
  )


def federated_secure_select(client_keys, max_key, server_val, select_fn):
  """Sends privately-selected values from a server database to clients.

  Args:
    client_keys: `federated_language.CLIENTS`-placed one-dimensional fixed-size
      non-negative `int32` keys used to select values from `database` to load
      for each client.
    max_key: A `federated_language.SERVER`-placed `int32` which is guaranteed to
      be greater than any of `client_keys`. Lower values may permit more
      optimizations.
    server_val: `federated_language.SERVER`-placed value used as an input to
      `select_fn`.
    select_fn: A `federated_language.Computation` which accepts unplaced
      `server_val` and a `int32` client key and returns a value to be sent to
      the client. `select_fn` should be deterministic (nonrandom).

  Returns:
    `federated_language.CLIENTS`-placed sequences of values returned from
    `select_fn`. In each
    sequence, the order of values will match the order of keys in the
    corresponding `client_keys` tensor. For example, a client with keys
    `[1, 2, ...]` will receive a sequence of values
    `[select_fn(server_val, 1), select_fn(server_val, 2), ...]`.

  Raises:
    TypeError: If `client_keys` is not of type `{int32[N]}@CLIENTS`, if
      `max_key` is not of type `int32@SERVER`, if `server_val` is not a
      server-placed value (`S@SERVER`), or if `select_fn` is not a function
      of type `<S, int32> -> RESULT`.
  """
  return _federated_select(
      client_keys, max_key, server_val, select_fn, secure=True
  )


def _federated_select(client_keys, max_key, server_val, select_fn, secure):
  """Internal helper for `federated_select` and `federated_secure_select`."""
  client_keys = value_impl.to_value(client_keys, type_spec=None)
  _check_select_keys_type(client_keys.type_signature, secure)
  max_key = value_impl.to_value(max_key, type_spec=None)
  expected_max_key_type = computation_types.FederatedType(
      np.int32, placements.SERVER
  )
  if not expected_max_key_type.is_assignable_from(max_key.type_signature):
    _select_parameter_mismatch(
        max_key.type_signature,
        'a 32-bit unsigned integer placed at server',
        'max_key',
        secure,
        expected_type=expected_max_key_type,
    )
  server_val = value_impl.to_value(server_val, type_spec=None)
  server_val = value_utils.ensure_federated_value(
      server_val, label='server_val'
  )
  expected_server_val_type = computation_types.FederatedType(
      computation_types.AbstractType('T'), placements.SERVER
  )
  if (
      not isinstance(server_val.type_signature, computation_types.FederatedType)
      or server_val.type_signature.placement is not placements.SERVER
  ):
    _select_parameter_mismatch(
        server_val.type_signature,
        'a value placed at server',
        'server_val',
        secure,
        expected_type=expected_server_val_type,
    )
  select_fn_param_type = computation_types.StructType([
      server_val.type_signature.member,  # pytype: disable=attribute-error
      np.int32,
  ])
  select_fn = value_impl.to_value(
      select_fn, type_spec=None, parameter_type_hint=select_fn_param_type
  )
  expected_select_fn_type = computation_types.FunctionType(
      select_fn_param_type, computation_types.AbstractType('U')
  )
  if not isinstance(
      select_fn.type_signature, computation_types.FunctionType
  ) or not select_fn.type_signature.parameter.is_assignable_from(  # pytype: disable=attribute-error
      select_fn_param_type
  ):
    _select_parameter_mismatch(
        select_fn.type_signature,
        'a function from state and key to result',
        'select_fn',
        secure,
        expected_type=expected_select_fn_type,
    )
  comp = building_block_factory.create_federated_select(
      client_keys.comp, max_key.comp, server_val.comp, select_fn.comp, secure
  )
  comp = _bind_comp_as_reference(comp)
  return value_impl.Value(comp)


def federated_secure_sum(value, max_input):
  """Computes a sum at `federated_language.SERVER` of a `value` placed on the `federated_language.CLIENTS`.

  This function computes a sum such that it should not be possible for the
  server to learn any clients individual value. The specific algorithm and
  mechanism used to compute the secure sum may vary depending on the target
  runtime environment the computation is compiled for or executed on. See
  https://research.google/pubs/pub47246/ for more information.

  Not all executors support `federated_language.federated_secure_sum()`; consult
  the
  documentation for the specific executor or executor stack you plan on using
  for the specific of how it's handled by that executor.

  The `max_input` argument is the maximum value (inclusive) that may appear in
  `value`. *Lower values may allow for improved communication efficiency.*
  Attempting to return a `value` higher than `max_input` is invalid, and will
  result in a failure at the given client.

  Example:

  ```python
  value = federated_language.federated_value(1, federated_language.CLIENTS)
  result = federated_language.federated_secure_sum(value, 1)

  value = federated_language.federated_value((1, 2), federated_language.CLIENTS)
  result = federated_language.federated_secure_sum(value, (1, 2))
  ```

  NOTE: To sum non-integer values or to sum integers with fewer constraints and
  weaker privacy properties, consider using `federated_sum`.

  Args:
    value: An integer or nested structure of integers placed at
      `federated_language.CLIENTS`, in the range `[0, max_input]`.
    max_input: A Python integer or nested structure of integers matching the
      structure of `value`. If integer `max_value` is used with a nested
      `value`, the same integer is used for each tensor in `value`.

  Returns:
    A representation of the sum of the member constituents of `value` placed
    on the `federated_language.SERVER`.

  Raises:
    TypeError: If the argument is not a federated value placed at
      `federated_language.CLIENTS`.
  """
  value = value_impl.to_value(value, type_spec=None)
  value = value_utils.ensure_federated_value(
      value, placements.CLIENTS, 'value to be summed'
  )
  type_analysis.check_is_structure_of_integers(value.type_signature)
  max_input_value = value_impl.to_value(max_input, type_spec=None)
  value_member_type = value.type_signature.member  # pytype: disable=attribute-error
  max_input_type = max_input_value.type_signature
  if not type_analysis.is_single_integer_or_matches_structure(
      max_input_type, value_member_type
  ):
    raise TypeError(
        'Expected `federated_secure_sum` parameter `max_input` to match '
        'the structure of `value`, with one integer max per tensor in '
        '`value`. Found `value` of `{}` and `max_input` of `{}`.'.format(
            value_member_type, max_input_type
        )
    )
  if isinstance(max_input_type, computation_types.TensorType) and isinstance(
      value_member_type, computation_types.StructType
  ):
    max_input_value = value_impl.to_value(
        structure.map_structure(lambda _: max_input, value_member_type),
        type_spec=None,
    )
  comp = building_block_factory.create_federated_secure_sum(
      value.comp, max_input_value.comp
  )
  comp = _bind_comp_as_reference(comp)
  return value_impl.Value(comp)


def federated_secure_sum_bitwidth(value, bitwidth):
  """Computes a sum at `federated_language.SERVER` of a `value` placed on the `federated_language.CLIENTS`.

  This function computes a sum such that it should not be possible for the
  server to learn any clients individual value. The specific algorithm and
  mechanism used to compute the secure sum may vary depending on the target
  runtime environment the computation is compiled for or executed on. See
  https://research.google/pubs/pub47246/ for more information.

  Not all executors support
  `federated_language.federated_secure_sum_bitwidth()`; consult the
  documentation for the specific executor or executor stack you plan on using
  for the specific of how it's handled by that executor.

  The `bitwidth` argument represents the bitwidth of the aggregand, that is the
  bitwidth of the input `value`. The federated secure sum bitwidth (i.e., the
  bitwidth of the *sum* of the input `value`s over all clients) will be a
  function of this bitwidth and the number of participating clients.

  Example:

  ```python
  value = federated_language.federated_value(1, federated_language.CLIENTS)
  result = federated_language.federated_secure_sum_bitwidth(value, 2)

  value = federated_language.federated_value([1, 1], federated_language.CLIENTS)
  result = federated_language.federated_secure_sum_bitwidth(value, [2, 4])

  value = federated_language.federated_value([1, [1, 1]],
  federated_language.CLIENTS)
  result = federated_language.federated_secure_sum_bitwidth(value, [2, [4, 8]])
  ```

  NOTE: To sum non-integer values or to sum integers with fewer constraints and
  weaker privacy properties, consider using `federated_sum`.

  Args:
    value: An integer value of a federated type placed at the
      `federated_language.CLIENTS`, in the range [0, 2^bitwidth - 1].
    bitwidth: An integer or nested structure of integers matching the structure
      of `value`. If integer `bitwidth` is used with a nested `value`, the same
      integer is used for each tensor in `value`.

  Returns:
    A representation of the sum of the member constituents of `value` placed
    on the `federated_language.SERVER`.

  Raises:
    TypeError: If the argument is not a federated value placed at
      `federated_language.CLIENTS`.
  """
  value = value_impl.to_value(value, type_spec=None)
  value = value_utils.ensure_federated_value(
      value, placements.CLIENTS, 'value to be summed'
  )
  type_analysis.check_is_structure_of_integers(value.type_signature)
  bitwidth_value = value_impl.to_value(bitwidth, type_spec=None)
  value_member_type = value.type_signature.member  # pytype: disable=attribute-error
  bitwidth_type = bitwidth_value.type_signature
  if not type_analysis.is_single_integer_or_matches_structure(
      bitwidth_type, value_member_type
  ):
    raise TypeError(
        'Expected `federated_secure_sum_bitwidth` parameter `bitwidth` to '
        'match the structure of `value`, with one integer bitwidth per tensor '
        'in `value`. Found `value` of `{}` and `bitwidth` of `{}`.'.format(
            value_member_type, bitwidth_type
        )
    )
  if isinstance(bitwidth_type, computation_types.TensorType) and isinstance(
      value_member_type, computation_types.StructType
  ):
    bitwidth_value = value_impl.to_value(
        structure.map_structure(lambda _: bitwidth, value_member_type),
        type_spec=None,
    )
  comp = building_block_factory.create_federated_secure_sum_bitwidth(
      value.comp, bitwidth_value.comp
  )
  comp = _bind_comp_as_reference(comp)
  return value_impl.Value(comp)


def sequence_map(fn, arg):
  """Maps a sequence `value` pointwise using a given function `fn`.

  This function supports two modes of usage:

  * When applied to a non-federated sequence, it maps individual elements of
    the sequence pointwise. If the supplied `fn` is of type `T->U` and
    the sequence `arg` is of type `T*` (a sequence of `T`-typed elements),
    the result is a sequence of type `U*` (a sequence of `U`-typed elements),
    with each element of the input sequence individually mapped by `fn`.
    In this mode of usage, `sequence_map` behaves like a computatation with type
    signature `<T->U,T*> -> U*`.

  * When applied to a federated sequence, `sequence_map` behaves as if it were
    individually applied to each member constituent. In this mode of usage, one
    can think of `sequence_map` as a specialized variant of `federated_map` that
    is designed to work with sequences and allows one to
    specify a `fn` that operates at the level of individual elements.
    Indeed, under the hood, when `sequence_map` is invoked on a federated type,
    it injects `federated_map`, thus
    emitting expressions like
    `federated_map(a -> sequence_map(fn, x), arg)`.

  Args:
    fn: A mapping function to apply pointwise to elements of `arg`.
    arg: A value of a type that is either a sequence, or a federated sequence.

  Returns:
    A sequence with the result of applying `fn` pointwise to each
    element of `arg`, or if `arg` was federated, a federated sequence
    with the result of invoking `sequence_map` on member sequences locally
    and independently at each location.

  Raises:
    TypeError: If the arguments are not of the appropriate types.
  """
  fn = value_impl.to_value(fn, type_spec=None)
  if not isinstance(fn.type_signature, computation_types.FunctionType):
    raise ValueError(
        'Expected `fn.type_signature` to be a `FunctionType`,'
        f' found {type(fn.type_signature)}.'
    )

  arg = value_impl.to_value(arg, type_spec=None)

  if isinstance(arg.type_signature, computation_types.SequenceType):
    comp = building_block_factory.create_sequence_map(fn.comp, arg.comp)
    comp = _bind_comp_as_reference(comp)
    return value_impl.Value(comp)
  elif isinstance(arg.type_signature, computation_types.FederatedType):
    parameter_type = computation_types.SequenceType(fn.type_signature.parameter)
    result_type = computation_types.SequenceType(fn.type_signature.result)
    intrinsic_type = computation_types.FunctionType(
        (fn.type_signature, parameter_type), result_type
    )
    intrinsic = building_blocks.Intrinsic(
        intrinsic_defs.SEQUENCE_MAP.uri, intrinsic_type
    )
    intrinsic_impl = value_impl.Value(intrinsic)
    local_fn = value_utils.get_curried(intrinsic_impl)(fn)
    return federated_map(local_fn, arg)
  else:
    raise TypeError(
        'Cannot apply `federated_language.sequence_map()` to a value of'
        ' type {}.'.format(arg.type_signature)
    )


def sequence_reduce(value, zero, op):
  """Reduces a sequence `value` given a `zero` and reduction operator `op`.

  This method reduces a set of elements of a sequence `value`, using a given
  `zero` in the algebra (i.e., the result of reducing an empty sequence) of some
  type `U`, and a reduction operator `op` with type signature `(<U,T> -> U)`
  that incorporates a single `T`-typed element of `value` into the `U`-typed
  result of partial reduction. In the special case of `T` equal to `U`, this
  corresponds to the classical notion of reduction of a set using a commutative
  associative binary operator. The generalized reduction (with `T` not equal to
  `U`) requires that repeated application of `op` to reduce a set of `T` always
  yields the same `U`-typed result, regardless of the order in which elements
  of `T` are processed in the course of the reduction.

  One can also invoke `sequence_reduce` on a federated sequence, in which case
  the reductions are performed pointwise; under the hood, we construct an
  expression  of the form
  `federated_map(x -> sequence_reduce(x, zero, op), value)`. See also the
  discussion on `sequence_map`.

  NOTE: When applied to a federated value this function does the reduce
  point-wise.

  Args:
    value: A value that is either a sequence, or a federated sequence.
    zero: The result of reducing a sequence with no elements.
    op: An operator with type signature `(<U,T> -> U)`, where `T` is the type of
      the elements of the sequence, and `U` is the type of `zero` to be used in
      performing the reduction.

  Returns:
    The `U`-typed result of reducing elements in the sequence, or if the `value`
    is federated, a federated `U` that represents the result of locally
    reducing each member constituent of `value`.

  Raises:
    TypeError: If the arguments are not of the types specified above.
  """
  value = value_impl.to_value(value, type_spec=None)
  zero = value_impl.to_value(zero, type_spec=None)
  op = value_impl.to_value(op, type_spec=None)
  # Check if the value is a federated sequence that should be reduced
  # under a `federated_map`.
  if isinstance(value.type_signature, computation_types.FederatedType):
    value_member_type = value.type_signature.member
    if not isinstance(value_member_type, computation_types.SequenceType):
      raise ValueError(
          'Expected a `federated_language.SequenceType`, found'
          f' {value_member_type}.'
      )
    zero_member_type = zero.type_signature.member  # pytype: disable=attribute-error
    ref_type = computation_types.StructType(
        [value_member_type, zero_member_type]
    )
    ref = building_blocks.Reference('arg', ref_type)
    arg1 = building_blocks.Selection(ref, index=0)
    arg2 = building_blocks.Selection(ref, index=1)
    call = building_block_factory.create_sequence_reduce(arg1, arg2, op.comp)
    fn = building_blocks.Lambda(ref.name, ref.type_signature, call)
    fn_value_impl = value_impl.Value(fn)
    args = building_blocks.Struct([value.comp, zero.comp])
    return federated_map(fn_value_impl, args)
  elif isinstance(value.type_signature, computation_types.SequenceType):
    comp = building_block_factory.create_sequence_reduce(
        value.comp, zero.comp, op.comp
    )
    comp = _bind_comp_as_reference(comp)
    return value_impl.Value(comp)
  else:
    raise NotImplementedError(
        f'Unexpected type found: {type(value.type_signature)}.'
    )


def sequence_sum(value):
  """Computes a sum of elements in a sequence.

  Args:
    value: A value of a type that is either a sequence, or a federated sequence.

  Returns:
    The sum of elements in the sequence. If the argument `value` is of a
    federated type, the result is also of a federated type, with the sum
    computed locally and independently at each location (see also a discussion
    on `sequence_map` and `sequence_reduce`).

  Raises:
    TypeError: If the arguments are of wrong or unsupported types.
  """
  value = value_impl.to_value(value, type_spec=None)
  if isinstance(value.type_signature, computation_types.SequenceType):
    element_type = value.type_signature.element
  else:
    if not isinstance(value.type_signature, computation_types.FederatedType):
      raise ValueError(
          'Expected `value.type_signature` to be a `FederatedType`,'
          f' found {type(value.type_signature)}.'
      )
    if not isinstance(
        value.type_signature.member, computation_types.SequenceType
    ):
      raise ValueError(
          'Expected `value.type_signature.member` to be a `SequenceType`,'
          f' found {type(value.type_signature.member)}.'
      )
    element_type = value.type_signature.member.element
  type_analysis.check_is_sum_compatible(element_type)

  if isinstance(value.type_signature, computation_types.SequenceType):
    comp = building_block_factory.create_sequence_sum(value.comp)
    comp = _bind_comp_as_reference(comp)
    return value_impl.Value(comp)
  elif isinstance(value.type_signature, computation_types.FederatedType):
    intrinsic_type = computation_types.FunctionType(
        value.type_signature.member, value.type_signature.member.element
    )
    intrinsic = building_blocks.Intrinsic(
        intrinsic_defs.SEQUENCE_SUM.uri, intrinsic_type
    )
    intrinsic_impl = value_impl.Value(intrinsic)
    return federated_map(intrinsic_impl, value)
  else:
    raise TypeError(
        'Cannot apply `federated_language.sequence_sum()` to a value of'
        ' type {}.'.format(value.type_signature)
    )



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definition of a federated computation."""

from typing import Optional

from federated_language.computation import computation_impl
from federated_language.computation import computation_wrapper
from federated_language.computation import function_utils
from federated_language.context_stack import context_stack_impl
from federated_language.federated_context import federated_computation_utils


def _federated_computation_wrapper_fn(
    fn,
    parameter_type,
    unpack: Optional[bool],
    name: Optional[str] = None,
    **kwargs
):
  """Wrapper function to plug orchestration logic into the framework."""
  del kwargs  # Unused.
  if parameter_type is None:
    parameter_name = None
  else:
    parameter_name = 'arg'
  fn = function_utils.wrap_as_zero_or_one_arg_callable(
      fn, parameter_type, unpack
  )
  context_stack = context_stack_impl.context_stack
  target_lambda, extra_type_spec = (
      federated_computation_utils.zero_or_one_arg_fn_to_building_block(
          fn,
          parameter_name,
          parameter_type,
          context_stack,
          suggested_name=name,
      )
  )
  return computation_impl.ConcreteComputation(
      computation_proto=target_lambda.to_proto(),
      context_stack=context_stack,
      annotated_type=extra_type_spec,
  )


federated_computation = computation_wrapper.ComputationWrapper(
    _federated_computation_wrapper_fn
)
federated_computation.__doc__ = """Decorates/wraps Python functions as federated/composite computations.

  The term *federated computation* as used here refers to any computation that
  uses API programming abstractions. Examples of such computations may include
  federated training or federated evaluation that involve both client-side and
  server-side logic and involve network communication. However, this
  decorator/wrapper can also be used to construct composite computations that
  only involve local processing on a client or on a server.

  The main feature that distinguishes *federated computation* function bodies
  in Python from the bodies of TensorFlow defuns is that whereas in the latter,
  one slices and dices `tf.Tensor` instances using a variety of TensorFlow ops,
  in the former one slices and dices `federated_language.Value` instances using operators.

  The supported modes of usage are identical to those for
  `federated_language.tensorflow.computation`.

  Example:

    ```python
    @federated_language.federated_computation((federated_language.FunctionType(np.int32, np.int32), np.int32))
    def foo(f, x):
      return f(f(x))
    ```

    The above defines `foo` as a function that takes a tuple consisting of an
    unary integer operator as the first element, and an integer as the second
    element, and returns the result of applying the unary operator to the
    integer twice. The body of `foo` does not contain federated communication
    operators, but we define it with `federated_language.federated_computation` as it can be
    used as building block in any section of code (except inside sections
    of pure TensorFlow logic).

  Args:
    *args: Either a Python function, or Federated Language type spec, or both
      (function first), or neither. See also
      `federated_language.tensorflow.computation` for an extended documentation.

  Returns:
    If invoked with a function as an argument, returns an instance of a
    computation constructed based on this function. If called without one, as
    in the typical decorator style of usage, returns a callable that expects
    to be called with the function definition supplied as a parameter. See
    also `federated_language.tensorflow.computation` for an extended documentation.
  """
